import gzip
import zlib

from io import StringIO, BytesIO, TextIOWrapper
from queue import Queue
from urllib import request
from typing import Any, Callable, Iterable, Union, Mapping, Sequence, Tuple, Iterator

from coba.exceptions import CobaException
from coba.primitives import Source, Filter
from coba.utilities  import try_else
from coba.pipes.utilities import resolve_params

class SourceFilters(Source):
    def __init__(self, *pipes: Union[Source,Filter]) -> None:
        self._pipes = sum((try_else(lambda: list(p),[p]) for p in pipes),[])

    @property
    def params(self) -> Mapping[str,Any]:
        return resolve_params(list(self))

    def read(self) -> Any:
        item = self._pipes[0].read()
        for filter in self._pipes[1:]:
            item = filter.filter(item)
        return item

    def __str__(self) -> str:
        return " | ".join(map(str,self._pipes))

    def __getitem__(self, index:int) -> Union[Source,Filter]:
        return self._pipes[index]

    def __iter__(self) -> Iterator[Union[Source,Filter]]:
        return iter(self._pipes)

    def __len__(self) -> int:
        return len(self._pipes)

class NullSource(Source[Any]):
    """A source which always returns an empty list."""

    def read(self) -> Iterable[Any]:
        return []

class IdentitySource(Source[Any]):
    """A source that reads from an iterable."""

    def __init__(self, item: Any, params: Mapping[str,Any] = None):
        """Instantiate an IterableSource.

        Args:
            item: The item to return from read.
            params: Teh params descirbing the source.
        """
        self._item = item
        self._params = params or {}

    @property
    def params(self) -> Mapping[str, Any]:
        return self._params

    def read(self) -> Iterable[Any]:
        return self._item

class NextSource(Source[Any]):

    def __init__(self, source: Source[Iterable[Any]]) -> None:
        self._source = source

    def read(self) -> Any:
        items = self._source.read()
        try:
            return next(iter(items))
        finally:
            if hasattr(items,'close') and callable(items.close):
                items.close()

class DiskSource(Source[Iterable[str]]):
    """A source that reads a file from disk.

    This source supports reading both plain text files as well gz compressed file.
    In order to make this distinction gzip files must end with a gz extension.
    """

    def __init__(self, path:str, mode:str='rt+', start_loc:int = 0, include_loc: bool = False):
        """Instantiate a DiskSource.

        Args:
            filename: The path to the file to read.
            mode: The mode with which the file should be read.
        """

        self._path        = path
        self._mode        = mode
        self._start_loc   = start_loc
        self._include_loc = include_loc

    def read(self) -> Union[Iterable[str],Iterable[Tuple[int,str]]]:
        #this stackoverflow question raises a potential problem
        #with this implementation https://stackoverflow.com/q/15353220/1066291

        #and this stackoverflow question points out a shortcoming
        #with this implementation #https://stackoverflow.com/a/59168992/1066291

        opener = gzip.open if ".gz" in self._path else open
        with opener(self._path, self._mode) as f:
            f.seek(self._start_loc)
            loc,line = f.tell(), f.readline()
            while line != '':
                line = line.rstrip('\r\n')
                yield (loc,line) if self._include_loc else line
                loc,line = f.tell(), f.readline()

class QueueSource(Source[Iterable[Any]]):
    """A source that reads from a queue."""

    def __init__(self, queue:Queue, block:bool=True, poison:Any=None) -> None:
        """Instantiate a QueueSource.

        Args:
            queue: The queue that should be read.
            block: Indicates if the queue should block when it is empty.
            poison: The poison pill that indicates when to stop blocking (if blocking).
        """
        self._queue    = queue or Queue()
        self._poison   = poison
        self._block    = block
        self._poisoned = False

    def read(self) -> Iterable[Any]:
        try:
            while self._block or self._queue.qsize() > 0:
                item = self._queue.get()

                if self._block and item == self._poison:
                    self._poisoned = True
                    break

                yield item
        except (EOFError,BrokenPipeError,TypeError):
            pass

from http.client import HTTPResponse

class HttpSource(Source[Union[str,Iterable[str]]]):
    """Get content from a web URL."""

    def __init__(self, url: str, chunk_size=None, timeout:float=None) -> None:
        """Instantiate an HttpSource.

        Args:
            url: Location we should get an HTTP response from.
            chunk_size: The number of bytes to read at a time.
            timeout: Seconds to wait for a response before failing.
        """
        self._url        = url
        self._chunk_size = chunk_size
        self._timeout    = timeout

    @staticmethod
    def _byte_io_(encoding:str, charset:str, bites: BytesIO) -> StringIO:
        # In my testing TextIOWrapper is a little slower than
        # codecs.getreader(charset). That said the format of
        # new lines in codecs is inconsistent across platforms
        # making unit testing problematic. Additionally, Python,
        # sees TextIOWrapper as a pure upgrade codecs.getreader.
        # https://peps.python.org/pep-0400/
        bites = gzip.open(bites) if encoding else bites
        return TextIOWrapper(bites,encoding=charset)

    @staticmethod
    def _resp_io_(resp: HTTPResponse) -> StringIO:
        encoding = resp.headers.get('Content-Encoding')
        charset  = next(iter(resp.info().get_charsets()),None) or 'utf-8'
        return HttpSource._byte_io_(encoding,charset,resp)

    @staticmethod
    def _byte_it_(encoding:str, charset:str, chunk:int, bites: BytesIO) -> Union[str,Iterable[str]]:
        # this is faster than the _byte_io_ pattern but we don't have
        # as many guardrails because we're rolling a lot of our own code
        if encoding == 'deflate':
            decomp = zlib.decompressobj(-zlib.MAX_WBITS).decompress
        elif encoding == "gzip":
            decomp = zlib.decompressobj(16+zlib.MAX_WBITS).decompress
        else:
            decomp = lambda x: x

        if not chunk:
            with bites as b:
                return decomp(b.read()).decode(charset)
        else:
            def chunks(decomp,charset,size,bites):
                with bites as b:
                    while chunk := b.read(size):
                        yield decomp(chunk).decode(charset)

            return DelimSource(IterableSource(chunks(decomp,charset,chunk,bites))).read()

    @staticmethod
    def _resp_it_(resp: HTTPResponse, chunk:int = None) -> Union[str,Iterable[str]]:
        encoding = resp.headers.get('Content-Encoding')
        charset  = next(iter(resp.info().get_charsets()),None) or 'utf-8'
        return HttpSource._byte_it_(encoding,charset,chunk,resp)

    def read(self) -> Union[str,Iterable[str]]:
        try:
            req  = request.Request(self._url,headers={'Accept-Encoding':'gzip, deflate'})
            return self._resp_it_(request.urlopen(req, timeout=self._timeout), self._chunk_size)
        except request.HTTPError as e:
            raise request.HTTPError(e.url, e.code, e.msg, e.hdrs, self._resp_io_(e.fp))

class DelimSource(Source[Iterable[str]]):
    """Chunk an Iterable[str] by delimeter."""

    def __init__(self, source: Source[Iterable[str]], delimeter:str = None) -> None:
        self._source = source
        self._delim  = delimeter

    def read(self) -> Iterable[str]:
        pending     = None
        split_lines = not self._delim
        delim       = self._delim

        if split_lines:
            for text in filter(None,self._source.read()):
                lines = text.splitlines()
                if pending:
                    lines[0] = pending + lines[0]
                    pending = None
                if text[-1] not in '\r\n':
                    pending = lines.pop()
                yield from lines
        else:
            for text in filter(None,self._source.read()):
                lines = text.split(delim)
                if pending:
                    lines[0] = pending + lines[0]
                    pending = None
                if text[-1] != delim:
                    pending = lines.pop()
                else:
                    lines.pop() #this makes the delim case behave similar to splitlines

                yield from lines

        if pending is not None:
            yield pending

class IterableSource(Source[Iterable[Any]]):
    """A source that reads from an iterable."""

    def __init__(self, iterable: Iterable[Any]=None):
        """Instantiate an IterableSource.

        Args:
            iterable: The iterable we should read from.
        """
        self.iterable = [] if iterable is None else iterable

    def read(self) -> Iterable[Any]:
        yield from self.iterable

class ListSource(Source[Sequence[Any]]):
    """A source that reads from a list."""

    def __init__(self, sequence: Sequence[Any]=None):
        """Instantiate a ListSource.

        Args:
            list: The sequence to read.
        """
        self._items = [] if sequence is None else sequence

    @property
    def items(self) -> Sequence[Any]:
        return self._items

    def read(self) -> Sequence[Any]:
        return self._items

class LambdaSource(Source[Any]):
    """A source that reads from a callable method."""

    def __init__(self, read: Callable[[],Any]):
        """Instantiate a LambdaSource.

        Args:
            read: A function to call for a return value when reading.
        """
        self._read = read

    def read(self) -> Iterable[Any]:
        return self._read()

class UrlSource(Source[Iterable[str]]):
    """A source that reads from a url.

    If the given url uses a file scheme or is a local path then
    a DiskSource is used internally.If the given url uses an http
    or https scheme then an HttpSource is used internally.
    """

    def __init__(self, url:str) -> None:
        """Instantiate a UrlSource.

        Args:
            url: The url to a resource. Can be either a web request or a local path.
        """
        self._url = url

        if url.startswith("http://") or url.startswith("https://"):
            self._source = HttpSource(url)
        elif url.startswith("file://"):
            self._source = DiskSource(url[7:])
        elif "://" not in url:
            self._source = DiskSource(url)
        else:
            raise CobaException("Unrecognized scheme, supported schemes are: http, https or file.")

    def read(self) -> Iterable[str]:
        return self._source.read()

class DataFrameSource(Source[Iterable[Mapping[str,Any]]]):

    def __init__(self, df) -> None:
        self._df = df

    def read(self) -> Iterable[Mapping[str,Any]]:
        "Iterate over DataFrame rows as dictionaries."
        yield from self._df.to_dict(orient='records')



import gzip

from queue import Queue
from itertools import islice
from collections import abc
from typing import Callable, Any, List, Union, Sequence, Iterable, Mapping, Iterator

from coba.primitives import Sink, Filter
from coba.utilities import try_else
from coba.pipes.utilities import resolve_params

class FiltersSink(Sink):
    def __init__(self, *pipes: Union[Filter,Sink]) -> None:
        self._pipes = sum((try_else(lambda: list(p),[p]) for p in pipes),[])

    @property
    def params(self) -> Mapping[str,Any]:
        return resolve_params(list(self))

    def write(self, item: Any):
        for filter in self._pipes[:-1]:
            item = filter.filter(item)
        self._pipes[-1].write(item)

    def __str__(self) -> str:
        return " | ".join(map(str,self._pipes))

    def __getitem__(self, index: int) -> Union[Filter,Sink]:
        return self._pipes[index]

    def __iter__(self) -> Iterator[Filter]:
        return iter(self._pipes)

    def __len__(self) -> int:
        return len(self._pipes)

class NullSink(Sink[Any]):
    """A sink which does nothing with written items."""
    def write(self, item: Any) -> None:
        pass

class ConsoleSink(Sink[Any]):
    """A sink which prints written items to console."""

    def write(self, item: Any) -> None:
        print(item)

class DiskSink(Sink[Union[str,Sequence[str]]]):
    """A sink which writes to a file on disk.

    This sink supports writing in either plain text or as a gz compressed file.
    In order to make this distinction gzip files must end with a gz extension.
    """

    def __init__(self, filename:str, mode:str='a+', batch:int=None) -> None:
        """Instantiate a DiskSink.

        Args:
            filename: The path to the file to write.
            mode: The mode with which the file should be written.
            batch: The number of lines to write before closing and reopening the file.
        """

        #Gzip compression performance is relative to batch size
        #see https://stackoverflow.com/a/18109797/1066291 for more info.

        self._filename = filename
        self._count    = 0
        self._file     = None
        self._mode     = mode
        self._batch    = batch

    def __enter__(self) -> 'DiskSink':
        self._count += 1

        if self._file is None:
            if ".gz" in self._filename:
                self._file = gzip.open(self._filename, f"{self._mode}b", compresslevel=6)
            else:
                self._file = open(self._filename, f"{self._mode}b")

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._count -= 1
        if self._count == 0 and self._file is not None:
            self._file.close()
            self._file = None

    def write(self, lines: Union[str,Iterable[str]]) -> None:
        if isinstance(lines,str):
            lines = [lines]

        lines = iter(lines)
        batch = None

        while self._unfinished(batch):
            batch = self._get_batch(lines)
            with self:
                for line in batch:
                    self._file.write((line + '\n').encode('utf-8'))
                    self._file.flush()

    def _get_batch(self, lines: Iterable[str]) -> Iterable[str]:
        batch = islice(lines,self._batch)
        if self._batch: batch = list(batch)
        return batch

    def _unfinished(self, batch:Any) -> bool:
        not_started  = batch is None
        not_finished = isinstance(batch,list) and len(batch) == self._batch
        return not_started or not_finished

class ListSink(Sink[Any]):
    """A sink which appends written items to a list."""

    def __init__(self, items: List[Any] = None, foreach: bool=False) -> None:
        """Instantiate a ListSink.

        Args:
            items: The list we wish to write to.
        """
        self.items = items if items is not None else []
        self._foreach = foreach

    def write(self, item) -> None:
        item = (item if self._foreach else [item])
        for i in item:
            self.items.append(list(i) if isinstance(i, abc.Iterator) else i)

class QueueSink(Sink[Any]):
    """A sink which puts written items into a Queue."""

    def __init__(self, queue:Queue=None, foreach:bool=False) -> None:
        """Instantiate a QueueSink.

        Args:
            queue: The queue to put written items into.
            foreach: Indicates whether to foreach over given items when writing.
        """
        self._queue   = queue or Queue()
        self._foreach = foreach

    def write(self, item: Any) -> None:
        try:
            item = (item if self._foreach else [item])
            for i in item:
                self._queue.put(i)
        except (EOFError,BrokenPipeError,AssertionError):
            pass

class LambdaSink(Sink[Any]):
    """A sink which passes written items to a callable function."""

    def __init__(self, write: Callable[[Any],None]):
        """Instantiate a LambdaSink.

        Args:
            write: A callable function written items will be passed to.
        """
        self._write = write

    def write(self, item: Any) -> None:
        return self._write(item)



import warnings
import importlib.util

from math import isfinite
from itertools import chain, islice, groupby
from collections import defaultdict
from typing import TypeVar, Iterable, Tuple, Union, Sequence, Any, Callable, Hashable

from coba.exceptions import CobaExit

def coba_exit(message:str):
    #we ignore warnings before exiting in order to make jupyter's output a little cleaner
    warnings.filterwarnings("ignore",message="To exit: use 'exit', 'quit', or Ctrl-D.")
    raise CobaExit(message) from None

class PackageChecker:
    @staticmethod
    def matplotlib(caller_name:str = None, strict:bool = True) -> None:
        """Raise ImportError with detailed error message if matplotlib is not installed.

        Functionality requiring matplotlib should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires matplotlib.
        """
        return PackageChecker._check(caller_name, "matplotlib.pyplot", "matplotlib", strict=strict)

    @staticmethod
    def vowpalwabbit(caller_name:str = None, strict:bool = True) -> None:
        """Raise ImportError with detailed error message if vowpalwabbit is not installed.

        Functionality requiring vowpalwabbit should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires vowpalwabbit.
        """
        return PackageChecker._check(caller_name, "vowpalwabbit", strict=strict)

    @staticmethod
    def pandas(caller_name:str = None, strict:bool = True) -> None:
        """Raise ImportError with detailed error message if pandas is not installed.

        Functionality requiring pandas should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires pandas.
        """
        return PackageChecker._check(caller_name, "pandas", strict=strict)

    @staticmethod
    def numpy(caller_name:str = None, strict:bool = True) -> None:
        """Raise ImportError with detailed error message if numpy is not installed.

        Functionality requiring numpy should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires numpy.
        """
        return PackageChecker._check(caller_name, "numpy", strict=strict)

    @staticmethod
    def scipy(caller_name:str = None, strict:bool = True) -> None:
        """Raise ImportError with detailed error message if scipy is not installed.

        Functionality requiring scipy should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires scipy.
        """
        return PackageChecker._check(caller_name, "scipy", strict=strict)

    @staticmethod
    def sklearn(caller_name:str = None, strict:bool = True) -> None:
        """Raise ImportError with detailed error message if sklearn is not installed.

        Functionality requiring sklearn should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires sklearn.
        """
        return PackageChecker._check(caller_name, "sklearn", "scikit-learn", strict=strict)

    @staticmethod
    def torch(caller_name:str = None, strict:bool = True) -> None:
        """Raise ImportError with detailed error message if torch is not installed.

        Functionality requiring torch should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires torch.
        """
        return PackageChecker._check(caller_name, "torch", strict=strict)

    @staticmethod
    def cloudpickle(caller_name:str = None, strict:bool = True) -> None:
        """Raise ImportError with detailed error message if cloudpickle is not installed.

        Functionality requiring cloudpickle should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires torch.
        """
        return PackageChecker._check(caller_name, "cloudpickle", strict=strict)

    def _check(caller_name:str, module_name:str, pkg_name:str = None, strict:bool = True):
        """
        Remarks:
            This pattern borrows heavily from sklearn. As of 6/20/2020 sklearn code could be found
            at https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/__init__.py
        """

        try:
            #if there are submodules a ModuleNotFoundError can be produced
            module_found = importlib.util.find_spec(module_name) is not None
        except ModuleNotFoundError:
            module_found = False

        if module_found:
            return True
        elif strict:
            pkg_name = pkg_name or module_name
            coba_exit(f"ERROR: {caller_name} requires the {pkg_name} package. You can install this package via `pip install {pkg_name}`.")
        else:
            return False

class KeyDefaultDict(defaultdict):
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError( key )
        else:
            value = self.default_factory(key)
            self[key] = value
            return value

_T = TypeVar("_T")
def peek_first(items: Iterable[_T], n:int=1, reduce:bool=True) -> Tuple[Union[_T,Sequence[_T]], Iterable[_T]]:
    items = iter(items)
    first = list(islice(items,n))

    items = [] if not first and n>0 else chain(first,items)
    first = [] if n==0 else None if not first else first[0] if n==1 and reduce else first

    return first, items

def try_else(f:Callable[[],Any], default: Any) -> Any:
    try:
        return f()
    except:
        return default

def minimize(obj,precision=5):
    """Transform objects to minimize string representations.

    This converts python data types into formats that have minimum
    string representations. For example, integer floats are printed
    as '1.0' so we convert these to an int which is '1'. Floats are
    rounded to the given precision otherwise default. Without this
    floats will be printed with 16 digits.
    """

    P = 10**precision

    if hasattr(obj,'ndim'): #numpy/pytorch
        obj = obj.tolist()
    if isinstance(obj,float):
        obj = int(obj) if obj.is_integer() else round(obj*P)/P if isfinite(obj) else obj
    if isinstance(obj,tuple):
        obj = list(obj)
    if isinstance(obj,(list,dict)):
        def minobj(o):
            if isinstance(o,dict):
                o,kv = o.copy(),o.items()
            elif isinstance(o,(list,tuple)):
                o,kv = list(o),enumerate(o)
            else:
                return o
            for k,v in kv:
                if isinstance(v,float) and isfinite(v):
                    o[k] = int(v) if v.is_integer() else round(v*P)/P
                else:
                    o[k] = minobj(v)
            return o
        obj = minobj(obj)

    return obj

def grouper(items: Sequence[Any], key:Callable[[Any],Hashable]=None, val:Callable[[Any],Any]=None, sorted_: bool= False) -> Iterable[Tuple[Hashable,Iterable[Any]]]:

    if sorted_:
        for k, group in groupby(items,key=key):
            if val: yield (k, map(val,group))
            else  : yield (k, group)

    else:
        groups = defaultdict(list)
        for key,items in groupby(items,key=key):
            if val: groups[key].extend(map(val,items))
            else  : groups[key].extend(items)
        yield from groups.items()



from copy import deepcopy,copy
from collections import abc
from itertools import count, compress, chain, filterfalse, islice, repeat
from typing import Dict, Any, Union, Callable, Iterator, Sequence, Mapping, Iterable, Tuple, Literal

from coba.primitives import Dense_, Sparse_, Sparse, Dense, Categorical, Filter
from coba.utilities import peek_first

class LazyDense(Dense_):
    __slots__ = ('_row','_enc','headers','missing')

    def __init__(self, row: Union[Callable[[],Sequence],Sequence], enc: Sequence = None, headers: Mapping = None, missing: bool = None) -> None:
        self._row = row
        self._enc = enc

        if headers is not None: self.headers = headers
        if missing is not None: self.missing = missing

    def _load_or_get(self) -> Sequence:
        row = self._row
        if not callable(row): return row
        row = row()
        self._row = row
        return row

    def __getitem__(self, key: int):
        key = key if key.__class__ is int else self.headers[key]
        val = self._load_or_get()[key]
        enc = self._enc

        if not enc:
            return val
        else:
            try:
                return enc[key](val)
            except:
                if val in ['?','']: return None
                raise

    def __iter__(self) -> Iterator:
        if self._enc:
            return self._enc_all()
        else:
            return iter(self._load_or_get())

    def __len__(self) -> int:
        return len(self._load_or_get())

    def _enc_all(self)->Iterator:
        #placing this here gives us a slight perf boost
        #but is a bit of a conflation of functionality.
        #In theory this error checking logic with None
        #should be in an ARFF specific implementation.
        for e,v in zip(self._enc, self._load_or_get()):
            try:
                yield e(v)
            except:
                if v in ['?','']:
                    yield None
                else:
                    raise

class LazySparse(Sparse_):
    __slots__=('_row','_enc','_nsp','_fwd','_inv','missing')

    def __init__(self, row: Union[Callable[[],Mapping],Mapping], enc: Mapping = {}, nsp: set = set(), fwd: Mapping = {}, inv: Mapping = {}, missing:bool = None) -> None:
        self._row    = row
        self._enc    = enc
        self._nsp    = nsp
        self._fwd    = fwd
        self._inv    = inv
        if missing is not None: self.missing = missing

    def _load_or_get(self) -> Mapping:
        row = self._row
        if not callable(row): return row
        row = row()
        self._row = row
        return row

    def __getitem__(self, key: str):
        key = self._fwd.get(key,key)
        enc = self._enc

        try:
            val = self._load_or_get()[key]
        except KeyError:
            if key not in self._nsp: raise
            val = "0"

        if not enc:
            return val
        else:
            try:
                return enc.get(key,lambda x:x)(val)
            except:
                if val in ['?','']: return None
                raise

    def __iter__(self) -> Iterator:
        return iter(self.keys())

    def __len__(self) -> int:
        return len(self._load_or_get())

    def keys(self) -> abc.KeysView:
        hdr_inv_get = self._inv.__getitem__ if self._inv else lambda x:x
        return set(map(hdr_inv_get, self._load_or_get().keys() | self._nsp))

    def items(self) -> Sequence:
        inv = self._inv
        enc = self._enc
        row = self._load_or_get()

        if enc:
            return tuple(self._enc_items())
        elif inv:
            return tuple((inv.get(k,k), v) for k,v in row.items())
        else:
            return tuple(row.items())

    def _enc_items(self)->Iterator:
        enc   = self._enc
        inv   = self._inv
        raw   = self._load_or_get()
        items = chain(raw.items(), zip(self._nsp-raw.keys(), repeat("0")))

        for k,v in items:
            try:
                v = enc.get(k,lambda x:x)(v)
            except:
                if v in ['?','']: v = None
                else: raise

            yield (inv.get(k,k) if inv else k,v)

class SparseDense(Dense_):
    __slots__=('_values','_length')

    def __init__(self, values: Dict[int,Any], length:int) -> None:
        self._row = None
        self._values = values
        self._length = length

    def copy(self) -> 'SparseDense':
        return SparseDense(self._values.copy(), self._length)

    def __setitem__(self, key:int, value:Any):
        key = key if key >= 0 else key+self._length

        if key < 0 or key >= self._length:
            raise IndexError("list index out of range")

        self._values[key] = value

    def __getitem__(self, key: int):
        key = key if key >= 0 else key+self._length

        if key < 0 or key >= self._length:
            raise IndexError("list index out of range")

        return self._values.get(key,0)

    def __iter__(self) -> Iterator:
        sort = sorted(self._values.items())

        yield from repeat(0,sort[0][0])
        yield sort[0][1]

        for p,n in zip(sort,sort[1:]):
            yield from repeat(0,n[0]-p[0]-1)
            yield n[1]

        yield from repeat(0,self._length-sort[-1][0]-1)

    def __len__(self) -> int:
        return self._length

class HeadDense(Dense_):
    __slots__=('_row','headers')

    def __init__(self, row: Dense, headers: Mapping[str,int] = None) -> None:
        self._row    = row
        self.headers = headers

    def __getitem__(self, key: Union[str,int]):
        return self._row[key if key.__class__ is int else self.headers[key]]

    def __iter__(self) -> Iterator:
        return iter(self._row)

    def __len__(self) -> int:
        return len(self._row)

class HeadSparse(Sparse_):

    __slots__=('_row', '_fwd', '_inv')

    def __init__(self, row: Sparse, head_fwd: Mapping[str,str], head_inv: Mapping[str,str]) -> None:
        self._row = row
        self._fwd = head_fwd
        self._inv = head_inv

    def __getitem__(self, key: Union[str,int]):
        return self._row[self._fwd[key]]

    def __iter__(self) -> Iterator:
        return iter(self.keys())

    def __len__(self) -> int:
        return len(self._row)

    def keys(self) -> abc.KeysView:
        head_map_inv_get = self._inv.__getitem__
        return set(map(head_map_inv_get, self._row.keys()))

    def items(self) -> Sequence:
        head_map_inv_get = self._inv.__getitem__
        return tuple((head_map_inv_get(k),v) for k,v in self._row.items())

class HeadRows(Filter[Iterable[Union[Dense,Sparse]],Iterable[Union[Dense,Sparse]]]):

    def __init__(self, headers: Union[Sequence,Mapping]) -> None:
        if isinstance(headers, abc.Mapping):
            self._mapping = headers
        else:
            self._mapping = dict(zip(headers, count()))

    def filter(self, rows: Iterable[Union[Dense,Sparse]]) -> Iterable[Union[Dense,Sparse]]:
        first, rows = peek_first(rows)

        if isinstance(first, Dense):
            mapping  = self._mapping
            yield from (HeadDense(row, mapping) for row in rows)
        else:
            mapping     = self._mapping
            mapping_inv = {v:k for k,v in self._mapping.items()}
            yield from (HeadSparse(row, mapping, mapping_inv) for row in rows)

class EncodeDense(Dense_):
    __slots__=('_row','_encoders')

    def __init__(self, row: Dense, encoders: Sequence) -> None:
        self._row      = row
        self._encoders = encoders

    def __getitem__(self, key: Union[int,str]):
        return self._encoders[key](self._row[key])

    def __iter__(self) -> Iterator:
        return (e(v) for e,v in zip(self._encoders, self._row))

    def __len__(self) -> int:
        return len(self._encoders)

class EncodeSparse(Sparse_):
    __slots__=('_row','_enc','_nsp')

    def __init__(self, row: Sparse, encoders: Mapping, not_sparse: set) -> None:
        self._row = row
        self._enc = encoders
        self._nsp = not_sparse

    def __getitem__(self, key: Union[int,str]):
        try:
            return self._enc.get(key,lambda x:x)(self._row[key])
        except KeyError:
            if key in self._nsp: return self._enc.get(key, lambda x:x)("0")

    def __iter__(self) -> Iterator:
        return iter(self._row.keys() | self._nsp)

    def __len__(self) -> int:
        return len(self._row.keys() | self._nsp)

    def keys(self) -> abc.KeysView:
        return self._row.keys() | self._nsp

    def items(self) -> Sequence:
        t1 = tuple((k, self._enc.get(k,lambda x:x)(v)) for k,v in self._row.items())
        t2 = tuple((k, self._enc[k]("0")) for k in self._nsp-self._row.keys())
        return t1+t2

class EncodeRows(Filter[Iterable[Union[Dense,Sparse]],Iterable[Union[Dense,Sparse]]]):

    def __init__(self, encoders: Union[Callable,Sequence,Mapping]) -> None:
        self._encoders = encoders

    def filter(self, rows: Iterable[Union[Any,Dense,Sparse]]) -> Iterable[Union[Dense,Sparse]]:
        enc = self._encoders
        first, rows = peek_first(rows)

        if not rows: return []

        if isinstance(first,Dense):
            if isinstance(enc,abc.Mapping):
                if hasattr(first, 'headers'):
                    enc = [ enc.get(h, enc.get(i, lambda x:x)) for i,h in enumerate(first.headers) ]
                else:
                    enc = [ enc.get(i, lambda x:x)             for i   in range(len(first))        ]
            return ( EncodeDense(row, enc) for row in rows )

        elif isinstance(first,Sparse):
            if isinstance(enc, abc.Sequence): enc = dict(enumerate(enc))
            nsp = set()
            for k,v in enc.items():
                try:
                    if v('0')!=0: nsp.add(k)
                except: pass #pragma: no cover
            return ( EncodeSparse(row, enc, nsp) for row in rows )

class DropOne(Dense_):
    __slots__=('_row','_ind')

    def __init__(self, row: Dense, ind: int) -> None:
        self._row = row
        self._ind = ind

    def __getitem__(self, key: int):
        if key >= self._ind: key += 1
        return self._row[key]

    def __iter__(self) -> Iterator:
        row = self._row
        ind = self._ind
        return iter(chain(islice(row,ind),islice(row,ind+1,None)))

    def __len__(self) -> int:
        return len(self._row)-1

class KeepDense(Dense_):
    __slots__=('_row', '_map', '_sel', '_len','headers')
    def __init__(self, row: Dense, mapping: Mapping[Union[str,int],int], selects: Sequence, len: int, headers: Mapping,) -> None:
        self._row = row
        self._map = mapping
        self._sel = selects
        self._len = len
        if headers: self.headers = headers

    def __getitem__(self, key: Union[int,str]):
        return self._row[self._map.get(key,10000000)]

    def __iter__(self) -> Iterator:
        return iter(compress(self._row, self._sel))

    def __len__(self) -> int:
        return self._len

class DropSparse(Sparse_):
    __slots__=('_row','_drop_set')

    def __init__(self, row: Sparse, drop_set: set = None) -> None:
        self._row      = row
        self._drop_set = drop_set

    def _key_check(self,key):
        if key not in self._drop_set:
            return key
        raise KeyError(key)

    def __getitem__(self, key: Union[int,str]):
        return self._row[self._key_check(key)]

    def __iter__(self) -> Iterator:
        return iter(self._row.keys()-self._drop_set)

    def __len__(self) -> int:
        return len(self.keys())

    def keys(self) -> abc.KeysView:
        return self._row.keys() - self._drop_set

    def items(self) -> Sequence:
        drop = self._drop_set
        yield from (item for item in self._row.items() if item[0] not in drop)

class DropRows(Filter[Iterable[Union[Dense,Sparse]], Iterable[Union[Dense,Sparse]]]):
    """A filter which drops rows and columns from in table shaped data."""

    def __init__(self,
        drop_cols: Sequence[Union[str,int]] = [],
        drop_row: Callable[[Union[Sequence,Mapping]], bool] = None) -> None:
        """Instantiate a Drop filter.

        Args:
            drop_cols: Feature names which should be dropped.
            drop_row: A function that accepts a row and returns True if it should be dropped.
        """

        self._drop_cols = set(drop_cols)
        self._drop_row  = drop_row

    @staticmethod
    def make_drop_row_args(first, drop_cols) -> Tuple:
        if isinstance(first,Dense):
            try:
                selects = [ not any(i in drop_cols for i in I) for I in enumerate(first.headers) ]
                headers = first.headers.items()
                indexes = list(compress(range(len(first)), selects))
            except:
                selects = [ i not in drop_cols for i in range(len(first)) ]
                headers = []
                indexes = list(compress(range(len(first)), selects))

            mapping = {k:v for k,v in chain(enumerate(indexes),headers)}
            length  = len(indexes)

            if headers:
                external_indexes = dict(zip(indexes,count()))
                external_headers = { h: external_indexes[i] for h,i in headers if i in external_indexes }
            else:
                external_headers = None

            return mapping, selects, length, external_headers
        else:
            return set(drop_cols)

    def filter(self, rows: Iterable[Union[Dense,Sparse]]) -> Iterable[Union[Dense,Sparse]]:

        drop_cols   = self._drop_cols
        first, rows = peek_first(rows)

        rows = rows if not self._drop_row else filterfalse(self._drop_row, rows)

        if not drop_cols:
            return rows
        elif isinstance(first,Dense):
            mapping, selects, length, headers = DropRows.make_drop_row_args(first, drop_cols)
            return (KeepDense(row, mapping, selects, length, headers) for row in rows)
        else:
            drop_set = DropRows.make_drop_row_args(first, drop_cols)
            return (DropSparse(row, drop_set) for row in rows)

class LabelDense(Dense_):
    __slots__=('_row','_ind','_tipe')

    def __init__(self, row: Dense, ind: int, tipe: Literal['c','r','m']) -> None:
        self._row  = row
        self._ind  = ind
        self._tipe = tipe

    def __getitem__(self, key: Union[int,str]):
        return self._row[key]

    def __iter__(self) -> Iterator:
        return iter(self._row)

    def __len__(self) -> int:
        return len(self._row)

    @property
    def feats(self) -> Dense:
        return DropOne(self._row, self._ind)

    @property
    def label(self) -> Any:
        return self._row[self._ind]

    @property
    def tipe(self) -> Literal['c','r','m']:
        return self._tipe

    @property
    def labeled(self)-> Tuple[Dense,Any,Literal['c','r','m']]:
        return (DropOne(self._row, self._ind), self._row[self._ind], self._tipe)

class LabelSparse(Sparse_):
    __slots__=('_row', '_tipe', '_key')

    def __init__(self, row: Sparse, key: str, tipe: Literal['c','r','m']) -> None:
        self._row  = row
        self._key  = key
        self._tipe = tipe

    def __getitem__(self, key: str):
        try:
            return self._row[key]
        except KeyError:
            if key == self._key:
                return 0
            raise

    def __iter__(self) -> Iterator:
        return iter(self._row.keys() | {self._key})

    def __len__(self) -> int:
        return len(self._row.keys() | {self._key})

    def keys(self) -> abc.KeysView:
        return self._row.keys() | {self._key}

    def items(self) -> Sequence:
        key = self._key
        row = self._row
        items = tuple(row.items())
        if key not in row.keys():
            items += ((key,0),)
        return items

    @property
    def feats(self) -> Dense:
        return DropSparse(self._row, {self._key})

    @property
    def label(self) -> Any:
        return self[self._key]

    @property
    def tipe(self) -> Literal['c','r','m']:
        return self._tipe

    @property
    def labeled(self)-> Tuple[Sparse,Any,Literal['c','r','m']]:
        return (DropSparse(self._row, {self._key}), self[self._key], self._tipe)

class LabelRows(Filter[Iterable[Union[Dense,Sparse]],Iterable[Union[Dense,Sparse]]]):

    def __init__(self, label: Union[int,str], tipe: Literal['c','r','m']) -> None:
        self.label = label
        self.tipe  = tipe

    def filter(self, rows: Iterable[Union[Dense,Sparse]]) -> Iterable[Union[Dense,Sparse]]:
        label = self.label
        tipe  = self.tipe
        first, rows = peek_first(rows)

        if isinstance(first,Dense):
            ind = first.headers[label] if isinstance(label,str) else label
            return map(LabelDense, rows, repeat(ind), repeat(tipe))
        else:
            return map(LabelSparse, rows, repeat(label), repeat(tipe))

class EncodeCatRows(Filter[Iterable[Union[Any,Dense,Sparse]], Iterable[Union[Any,Dense,Sparse]]]):
    def __init__(self, tipe=Literal["onehot","onehot_tuple","string"]) -> None:
        self._tipe = tipe

    def filter(self, rows: Iterable[Union[Any,Dense,Sparse]]) -> Iterable[Union[Any,Dense,Sparse]]:
        if self._tipe is None: return rows
        first, rows = peek_first(rows)
        if not rows: return []

        if isinstance(first,(Dense,Sparse)):
            rows = self._encode_collection(rows, first)
        else:
            rows = self._encode_values(rows, first)

        return rows

    def _encode_values(self, rows, first):
        if not isinstance(first,Categorical):
            yield from rows
        elif self._tipe == 'string':
            yield from map(str,rows)
        else:
            yield from (r.as_onehot for r in rows)

    def _encode_collection(self, rows, first):
        get_string = 'string' == self._tipe
        flat_onehot = 'onehot' == self._tipe

        def catkey(o):
            if isinstance(o,dict):
                o,keys = o.copy(),o.keys()
            elif isinstance(o,(list,tuple)):
                o,keys = list(o),range(len(o)-1,-1,-1)
            else:
                return
            for k in keys:
                v = o[k]
                if isinstance(v,Categorical):
                    yield k
                else:
                    K = list(catkey(v))
                    if K: yield [k,K]
            return o

        def catset(o,k):
            #k is Tuple[key,list]
            if len(k) == 2 and isinstance(k[1],list):
                k,K = k
                row = o[k]
                row = list(row) if isinstance(row,tuple) else copy(row)
                o[k] = row
                catset(row,K)
            #k is list of keys
            else:
                if get_string:
                    for _k in k:
                        o[_k] = str(o[_k])
                elif flat_onehot:
                    for _k in k:
                        if isinstance(o,list):
                            h = o.pop(_k).as_onehot
                            n = len(o)
                            j = len(h)
                            z = _k-n
                            o.extend(h)
                            if z != 0:
                                o[_k:_k+j],o[z:] = o[n:],o[_k:n]
                        else:
                            h = o.pop(_k).as_onehot
                            for i,v in enumerate(h):
                                if i != 0: o[f'{_k}_{v}'] = i
                else:
                    for _k in k:
                        o[_k] =  o[_k].as_onehot

        catkeys = list(catkey(first))

        if not catkeys:
            yield from rows
        else:
            #cat_cols is list of numbers or list of lists
            is_nums = isinstance(catkeys[0],int)
            for row in rows:
                row = list(row) if isinstance(row,tuple) else copy(row)

                if is_nums:
                    catset(row,catkeys)
                else:
                    for k in catkeys: catset(row,k)

                yield row



"""This module contains functionality for defining dataflows and workflows within Coba.

This module is primarily intended for internal use within Coba. However, it is documented
and made public for advanced use cases where existing Environment and Experiment creation
functionality is not sufficient. That being said, a good understanding of the patterns
in coba.pipes can help one understand how to best take advantage what Coba has to offer.
"""

from coba.pipes.multiprocessing import Multiprocessor

from coba.pipes.filters import Take, Shuffle, Structure, Identity, Flatten, Default, Reservoir
from coba.pipes.filters import Encode, Cache, Slice, Insert, FiltersFilter

from coba.pipes.rows    import LabelRows, EncodeRows, HeadRows, DropRows, EncodeCatRows
from coba.pipes.rows    import LazyDense, EncodeDense, LabelDense, KeepDense, HeadDense
from coba.pipes.rows    import LazySparse, EncodeSparse, LabelSparse, DropSparse, SparseDense
from coba.pipes.readers import ManikReader, LibsvmReader, CsvReader, ArffReader
from coba.pipes.sources import NullSource, IdentitySource, DiskSource, IterableSource, DataFrameSource
from coba.pipes.sources import QueueSource, HttpSource, LambdaSource, UrlSource, ListSource, NextSource
from coba.pipes.sources import SourceFilters, DelimSource
from coba.pipes.sinks   import NullSink, ConsoleSink, DiskSink, ListSink, QueueSink, LambdaSink, FiltersSink
from coba.pipes.lines   import SourceSink, ThreadLine, ProcessLine

from coba.pipes.core import Pipes, Foreach, join



"""Global execution context for decoupled sharing of settings and information."""

from coba.context.cachers import Cacher, NullCacher, MemoryCacher, DiskCacher, ConcurrentCacher
from coba.context.loggers import Logger, NullLogger, BasicLogger, IndentLogger, ExceptionLogger, ExceptLog, NameLog, StampLog, DecoratedLogger
from coba.context.core    import CobaContext, ExperimentConfig



from math import isnan
from abc import ABC, abstractmethod
from typing import Sequence,Tuple,Callable,Literal

from coba.exceptions import CobaException
from coba.utilities import PackageChecker
from coba.statistics import mean, stdev, phi

class PointAndInterval(ABC):
    """Calculate a point estimate and a confidence interval."""

    @abstractmethod
    def point(self, sample: Sequence[float]) -> float:
        """Calculate point estimate of a statistic.

        Args:
            sample: Sample to calculate a statistic for.

        returns:
            A point estimate of a statistic.
        """
        ...

    @abstractmethod
    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
        """Calculate a point estimate and a confidence interval.

        Args:
            sample: Sample to calculate a statistic and its confidence interval.

        returns:
            A point estimate of a statistic along with a hi/lo confidence interval.
                The lo and hi values should be the size of the plotted error bars.
        """
        ...

class StdDevCI(PointAndInterval):
    """Calculate mean and standard deviation interval."""

    def point(self,sample: Sequence[float]) -> float:
        return mean(sample)

    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
        mu = mean(sample)
        sd = round(0 if len(sample) == 1 else stdev(sample),5)
        return (mu, (sd,sd))

class StdErrCI(PointAndInterval):
    """Calculate mean and standard error interval."""

    def __init__(self, confidence:float = .95) -> None:
        """Instantiate a StdErrCI.

        Args:
            confidence: The desired confidence level of the interval. Should be in [0,1).
        """

        bot = 0
        top = 10

        cdf_curr   = phi((bot+top)/2)
        cdf_target = .5+confidence/2

        while abs(cdf_curr - cdf_target) > .0001:
            if cdf_curr < cdf_target:
                bot = (bot+top)/2
            else:
                top = (bot+top)/2
            cdf_curr = phi((bot+top)/2)

        self._z_score = (bot+top)/2
        self._inner   = StdDevCI()

    def point(self,sample: Sequence[float]) -> float:
        return self._inner.point(sample)

    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
        (mu,(sd,sd)) = self._inner.point_interval(sample)
        sqrtn = len(sample)**.5
        se    = round(sd/sqrtn,5)
        ci    = self._z_score*se
        return (mu, (ci,ci))

class BootstrapCI(PointAndInterval):
    """Calculate a statistic and its Bootstrap CI."""

    def __init__(self, confidence:float, statistic:Callable[[Sequence[float]], float]) -> None:
        """Instantiate a BootstrapCI.

        Args:
            confidence: The desired confidence level of the interval. Should be in [0,1).
            statistic: A callable which returns a point estimate given a sample.
        """
        PackageChecker.scipy('BootstrapCI')
        self._stat = statistic
        self._args = dict(method='basic', vectorized=False, n_resamples=1000, random_state=1, confidence_level=confidence)

    def point(self,sample: Sequence[float]) -> float:
        return self._stat(sample)

    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
        from scipy.stats import bootstrap

        p   = self._stat(sample)

        if len(sample) < 3:
            l,h = p,p
        else:
            l,h = bootstrap([sample], self._stat, **self._args).confidence_interval

        return (p, (max(p-l,0),max(h-p,0)))

class BinomialCI(PointAndInterval):
    """Calculate the mean and interval of a binomial."""

    def __init__(self, method:Literal['wilson', 'clopper-pearson']):
        """Instantiate a BinomialCI.

        Args:
            method: The method to calculate the Binomial confidence interval
        """
        self._method = method

    def point(self,sample: Sequence[float]) -> float:
        return sum(sample)/len(sample)

    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
        if set(sample) - set([0,1]):
            raise CobaException("A binomial confidence interval can only be calculated on values of 0 and 1.")

        if self._method == "wilson":
            z_975 = 1.96 #z-score for .975 area to the left
            p_hat = sum(sample)/len(sample)
            n     = len(sample)
            Q     = z_975**2/(2*n)

            #https://www.itl.nist.gov/div898/handbook/prc/section2/prc241.htm
            interval_num = z_975*((p_hat*(1-p_hat))/n + Q/(2*n))**(.5)
            location_num = (p_hat+Q)

            interval_den = (1+2*Q)
            location_den = (1+2*Q)

            interval = interval_num/interval_den
            location = location_num/location_den

            lo = p_hat-(location-interval)
            hi = (location+interval)-p_hat

        else:
            PackageChecker.scipy("BinomialConfidenceInterval")
            from scipy.stats import beta

            lo = beta.ppf(.05/2, sum(sample), len(sample) - sum(sample) + 1)
            hi = beta.ppf(1-.05/2, sum(sample) + 1, len(sample) - sum(sample))
            p_hat = sum(sample)/len(sample)

            lo = 0.0 if isnan(lo) else lo
            hi = 1.0 if isnan(hi) else hi

            lo = p_hat-lo
            hi = hi-p_hat

        return (p_hat, (round(lo,5), round(hi,5)) )



from coba.primitives import Learner, Context, Actions, Action, Prob

class MisguidedLearner(Learner):
    """A learner trained on incorrect reward information.

    Remarks:
        This is useful to generate difficult logged data in off-policy experiments.
    """
    def __init__(self, learner: Learner, shifter:float, scaler:float) -> None:
        """Instantiate a MisguidedLearner.

        Args:
            learner: The learner that will be trained using incorrect rewards.
            shifter: A constant that will be added to all reward values during learning.
            scaler: A constant that will be multipled by the reward value during learning.

        Remarks:
            The final reward value that learner will see is shifter+scaler*reward.
        """
        self._learner = learner
        self._shifter = shifter
        self._scaler  = scaler

    @property
    def params(self):
        return {**self._learner.params, 'misguided': [self._shifter,self._scaler]}

    def score(self, context: 'Context', actions: 'Actions', action: 'Action') -> 'Prob':
        return self._learner.score(context, actions, action)

    def predict(self, context: 'Context', actions: 'Actions'):
        return self._learner.predict(context, actions)

    def learn(self, context: 'Context', action: 'Action', reward: float, probability: float, **kwargs):
        self._learner.learn(context, action, self._shifter + self._scaler*reward, probability,**kwargs)



"""Efficient Result loading, slicing, and analyzing """

from coba.results.core import Table, Result, TransactionDecode, TransactionEncode, TransactionResult, moving_average, Missing
from coba.results.errors import PointAndInterval, StdDevCI, StdErrCI, BootstrapCI, BinomialCI



import time
import warnings

from operator import mul
from statistics import mean
from bisect import insort

from typing import Any, Iterable, Sequence, Mapping, Optional, Literal

from coba.exceptions  import CobaException
from coba.random      import CobaRandom
from coba.context     import CobaContext
from coba.safety      import SafeLearner
from coba.primitives  import is_batch, Dense, Sparse, Learner, Environment, Evaluator
from coba.statistics  import percentile
from coba.utilities   import PackageChecker, peek_first

def get_ope_loss(learner) -> float:
    # OPE loss metric is only available for VW models
    try:
        return learner.learner._vw._vw.get_sum_loss()
    except AttributeError:
        return float("nan")

class SequentialCB(Evaluator):
    """Sequential evaluation for CB learners."""

    _ONLY_DISCRETE    = set()
    _IMPLICIT_EXCLUDE = {"context", "actions", "rewards", "action", "reward", "probability", "eval_rewards", "learn_rewards"}

    def __init__(self,
        record: Sequence[Literal['reward','time','probability','action','context','actions','rewards']] = ['reward','action','probability'],
        learn : Optional[Literal['on','off','ips','dr','dm']] = 'on',
        eval  : Optional[Literal['on','ips','dr','dm']] ='on',
        seed  : float = None) -> None:
        """Instantiate a SequentialCB evaluator.

        Args:
            record: Variables to record for each learner interaction.
            learn: action and reward revealed to learner on `learn` call:
                *on* --- on-policy action/reward (requires 'actions' and 'rewards'),
                *off* --- off-policy action/reward (requires 'action' and 'reward'),
                *ips* --- on-policy action/ips-reward (requires 'actions', 'action', 'reward', and 'probability'),
                *dr* --- on-policy action/dr-reward (requires 'actions', 'action', and 'reward'),
                *dm* --- on-policy action/dm-reward (requires 'actions', 'action', and 'reward'),
                *None* --- `learn` is not called.
            eval: reward recorded given learner prediction:
                *on* --- actual reward (requires 'actions' and 'rewards'),
                *ips* --- ips-reward (requires 'actions', 'action', 'reward', and 'probability'),
                *dr* --- dr-reward (requires 'actions', 'action', and 'reward'),
                *dm* --- dm-reward (requires 'actions', 'action', and 'reward'),
                *None* --- no reward is recorded.
            seed: Determine which action is played when learners return an action PMF.
        """

        self._record = [record] if isinstance(record,str) else record
        self._learn  = learn or ''
        self._eval   = eval or ''
        self._seed   = seed

        if 'ope_loss' in self._record:
            # OPE loss metric is only available for VW models
            # Divide by the number of samples for the average loss metric and see this article for more info
            # https://vowpalwabbit.org/docs/vowpal_wabbit/python/latest/tutorials/off_policy_evaluation.html
            PackageChecker.vowpalwabbit('OffPolicyEvaluator')

    @property
    def params(self) -> Mapping[str,Any]:
        return {'learn': self._learn, 'eval': self._eval, 'seed': self._seed }

    def _required(self, has_score:bool) -> set:
        learn,eval = self._learn,self._eval

        pred = (learn and learn != 'off') or (eval and (eval != 'ips' or not has_score))
        off  = (learn and learn != 'on')  or (eval and eval != 'on')
        rwds = (learn == 'on')            or (eval == 'on')

        required_keys = set()
        if pred: required_keys.update(['actions'])
        if off : required_keys.update(['action','reward'])
        if rwds: required_keys.update(['rewards'])

        return required_keys

    def _validate(self, first: Mapping, has_score:bool) -> None:
        required_keys = self._required(has_score)
        missing_keys  = required_keys-first.keys()

        if missing_keys:
            raise CobaException(f"{self.__class__.__name__}(learn={self._learn or None},eval={self._eval or None}) requires {missing_keys}.")

        if not self._discrete(first):
            for metric in set(self._record).intersection(SequentialCB._ONLY_DISCRETE):
                warnings.warn(f"The {metric} metric can only be calculated for discrete environments")#pragma: no cover

    def _discrete(self,interaction):
        if 'actions' not in interaction:
            return False
        else:
            batched   = interaction and (is_batch(interaction.get('context')) or is_batch(interaction.get('actions')))
            actions   = interaction['actions'][0] if batched else interaction['actions']
            discrete  = len(actions or []) > 0

        return discrete

    def _results(self,learner:SafeLearner,first:Mapping, interactions:Iterable[Mapping]) -> Iterable[Mapping[Any,Any]]:
        #import here to avoid circular dependencies...
        from coba.environments.filters import OpeRewards, Batch, BatchSafe

        batched   = first and (is_batch(first.get('context')) or is_batch(first.get('action')) or is_batch(first.get('actions')))
        discrete  = self._discrete(first)
        has_score = learner.has_score

        learn,eval = self._learn,self._eval

        N          = 0
        n_zero_lrn = 0
        n_zero_val = 0
        n_zeroes   = lambda R: int(R==0) if not batched else sum([r==0 for r in R])
        lrn_warned   = False
        val_warned   = False

        lrn_on  = learn == 'on'
        lrn_off = learn == 'off'
        lrn_ips = learn == 'ips'
        lrn_dr  = learn == 'dr'
        lrn_dm  = learn == 'dm'
        val_on  = eval  == 'on'
        val_ips = eval  == 'ips'
        val_dr  = eval  == 'dr'
        val_dm  = eval  == 'dm'

        has_context = 'context'     in first
        has_actions = 'actions'     in first
        has_rewards = 'rewards'     in first
        has_reward  = 'reward'      in first
        has_action  = 'action'      in first
        has_prob    = 'probability' in first

        out_prob     = 'probability' in self._record and eval
        out_time     = 'time'        in self._record
        out_action   = 'action'      in self._record and eval
        out_context  = 'context'     in self._record
        out_actions  = 'actions'     in self._record and has_actions
        out_rewards  = 'rewards'     in self._record and has_rewards
        out_reward   = 'reward'      in self._record and eval
        out_ope_loss = 'ope_loss'    in self._record and eval

        should_pred = (learn and not lrn_off) or val_on or val_dm or val_dr or (val_ips and not has_score) or out_action or out_prob

        if out_rewards and discrete:
            get_rewards = (lambda Rs,As:[[R(a) for a in A] for R,A in zip(Rs,As)]) if batched else (lambda R,A: [R(a) for a in A])
        if out_rewards and not discrete:
            get_rewards = lambda R,_: R

        info = CobaContext.learning_info
        info.clear()

        learn_type = 'IPS' if lrn_ips else 'DR' if lrn_dr else 'DM' if lrn_dm else None
        eval_type  = 'IPS' if val_ips else 'DR' if val_dr else 'DM' if val_dm else None

        if learn_type:
            interactions = BatchSafe(OpeRewards(learn_type,target='learn_rewards',features=[1,'a','xa'])).filter(interactions)

        if eval_type and eval_type != learn_type:
            interactions = BatchSafe(OpeRewards(eval_type,target='eval_rewards',features=[1,'a','xa'])).filter(interactions)

        learn_target = 'learn_rewards'
        eval_target  = 'eval_rewards' if eval_type and eval_type != learn_type else 'learn_rewards'

        for interaction in interactions:

            context = interaction['context'     ] if has_context else None
            actions = interaction['actions'     ] if has_actions else None
            rewards = interaction['rewards'     ] if has_rewards else None
            off_rwd = interaction['reward'      ] if has_reward  else None
            off_act = interaction['action'      ] if has_action  else None
            off_pr  = interaction['probability' ] if has_prob    else None

            lrn_rwds = interaction[learn_target] if learn_type else rewards if lrn_on else None
            val_rwds = interaction[eval_target ] if eval_type  else rewards if val_on else None

            N += 1 if not batched else len(lrn_rwds) if lrn_rwds else len(val_rwds)

            start = time.time()
            if should_pred: on_act,on_pr,on_kw=learner.predict(context,actions)
            pred_time = time.time()-start

            if eval:
                if val_ips and has_score and not should_pred:
                    scr_ = learner.score(context,actions,off_act)
                    rwd_ = val_rwds(off_act)
                    eval_reward = scr_*rwd_ if not batched else Batch.List(map(mul,scr_,rwd_))
                else:
                    eval_reward = val_rwds(on_act)

                n_zero_val += n_zeroes(eval_reward)

            if learn:
                learn_reward = off_rwd if lrn_off else lrn_rwds(on_act)
                n_zero_lrn += n_zeroes(learn_reward)
                start = time.time()
                if lrn_off: learner.learn(context, off_act, learn_reward, off_pr         )
                else      : learner.learn(context, on_act , learn_reward, on_pr , **on_kw)
                learn_time = time.time()-start

            if 200 > N and N > 20 and not lrn_warned and learn_type == 'IPS' and n_zero_lrn <= (N*.01):
                lrn_warned = True
                #The warning will display even in quiet mode due to CobaException
                CobaContext.logger.log(CobaException(
                    "WARNING: the learner's predicted actions are highly correlated"
                    " with the logging policy and we are learning from an IPS reward"
                    " estimate. This can cause biased learning due to IPS assuming"
                    " independence between logging and learning policies. To solve"
                    " this consider using DM or DR instead of IPS, make sure the"
                    " rng seed is not the same in the logging and learning policy,"
                    " or shuffle the order of interactions in the Environment."
                ))

            if 200 > N and N > 20 and not val_warned and eval_type == 'IPS' and not has_score and n_zero_val <= (N*.01):
                val_warned = True
                #The warning will display even in quiet mode due to CobaException
                CobaContext.logger.log(CobaException(
                    "WARNING: the learner's predicted actions are highly correlated"
                    " with the logging policy and we are evaluating from an IPS reward"
                    " estimate. This can cause biased evaluation due to IPS assuming"
                    " independence between logging and learning policies. To solve"
                    " this consider using DM or DR instead of IPS or make sure the"
                    " rng seed is not the same in the logging and learning policy,"
                    " or shuffle the order of interactions in the Environment."
                ))

            out = {}

            if out_time    : out['predict_time'] = pred_time
            if out_time    : out['learn_time']   = learn_time
            if out_context : out['context']      = context
            if out_actions : out['actions']      = actions
            if out_action  : out['action']       = on_act
            if out_reward  : out['reward']       = eval_reward
            if out_rewards : out['rewards']      = get_rewards(rewards,actions)
            if out_ope_loss: out['ope_loss']     = get_ope_loss(learner)

            if out_prob and should_pred and on_pr is not None:
                out['probability'] = on_pr

            out.update({k: interaction[k] for k in interaction.keys()-SequentialCB._IMPLICIT_EXCLUDE})

            if info:
                out.update(info)
                info.clear()

            if out:
                yield out

    def evaluate(self, environment: Optional[Environment], learner: Optional[Learner]) -> Iterable[Mapping[Any,Any]]:

        first, interactions = peek_first(environment.read())
        seed = self._seed if self._seed is not None else CobaContext.store.get("experiment_seed")

        if not interactions: return []

        from coba.environments import Finalize, Unbatch, BatchSafe

        learner = SafeLearner(learner, seed)
        self._validate(first,learner.has_score)
        results = self._results(learner, first, BatchSafe(Finalize()).filter(interactions))

        #We Unbatch to work with Result
        yield from Unbatch().filter(results)

class SequentialIGL(Evaluator):
    """Sequential evaluation for IGL learners."""

    def __init__(self,
        record: Sequence[Literal['reward','feedback','time','prob','action','context','actions','rewards','feedbacks']] = ['reward','feedback'],
        seed: int = None) -> None:
        """Instantiate SequentialIGL learners.

        Args:
            record: Variables to record for each learner interaction.
            seed: Determine which action is played when learners return an action PMF.
        """

        self._record = list(record)
        self._seed   = seed

    @property
    def params(self) -> Mapping[str,Any]:
        return { 'seed': self._seed }

    def evaluate(self, environment: Optional[Environment], learner: Optional[Learner]) -> Iterable[Mapping[Any,Any]]:

        class IglEnvironment:
            def __init__(self,env: Environment) -> None:
                self._env = env

            def read(self):
                first,interactions = peek_first(iter(self._env.read()))

                if not interactions:
                    return

                is_missing = 'context' not in first
                is_sparse  = not is_missing and isinstance(first['context'],Sparse)
                is_dense   = not is_missing and isinstance(first['context'],Dense)

                for interaction in interactions:

                    new = interaction.copy()

                    f,r = new['feedbacks'],new['rewards']
                    new['feedbacks'],new['rewards'] = r,f

                    userid = new['userid']

                    if is_missing:
                        new['context'] = userid
                    elif is_sparse:
                        new['context'] = dict(userid=userid,**new['context'])
                    elif is_dense:
                        new['context'] = (userid,)+tuple(new['context'])
                    else:
                        new['context'] = (userid, new['context'])

                    yield new

        envIGL = IglEnvironment(environment)
        record = self._record+['action']
        seed   = self._seed

        out_action   = 'action'   in self._record
        out_feedback = 'feedback' in self._record
        out_reward   = 'reward'   in self._record

        for out in SequentialCB(record,seed=seed).evaluate(envIGL,learner):
            if out_feedback  : out['feedback'] = out['reward']
            if out_reward    : out['reward']   = out.pop('feedbacks')(out['action'])
            if not out_action: del out['action']

            yield out

class RejectionCB(Evaluator):
    """Rejective evaluation for CB learners.

    This evaluator uses rejection sampling to simulate on-policy learner performance
    using only off-policy logged data. For this to work the evaluator requires each
    interaction to have 'actions', 'action', 'reward', and 'probability'.

    This gives an unbiased estimation of on-policy performance assuming two conditions

        1. The reward distribution of each interaction is stationary.
        2. The `cpct` parameter of the evaluator is set to 0.

    Remarks:
        This is an implementation of Dudík et al. (2012). The `cpct` parameter of our
        implementation is what Dudík calls *q* and `cinit` is Dudík calls `c1`. To use
        double-robust off-policy estimation as Dudík does also set `ope` to 'dr'.

    References:
        * Miroslav Dudík, Dumitru Erhan, John Langford, and Lihong Li. 2012. Sample-efficient
          nonstationary policy evaluation for contextual bandits. In Proceedings of the
          Twenty-Eighth Conference on Uncertainty in Artificial Intelligence (UAI'12). AUAI
          Press, Arlington, Virginia, USA, 247-254.
    """

    def __init__(self,
        record: Sequence[Literal['context','actions','action','reward','probability','time']] = ['reward'],
        ope   : Optional[Literal['ips','dr','dm']] = None,
        cpct  : float = .005,
        cmax  : float = 1.0,
        cinit : float = None,
        seed  : float = None) -> None:
        """Instantiate a RejectionCB evaluator.

        Args:
            record: The datapoints to record for each interaction.
            ope: Indicates whether off-policy estimates should be included from rejected training examples.
            cpct: The unbiased case is q = 0. Smaller values give better estimates but rejects more data.
            cmax: The maximum value that the evaluator is allowed to use for `c` (the rejection sampling multiplier).
                To get an unbiased estimate we need a `c` value such that c*on_prob/log_prob <= 1 for all
                on_prob/log_prob. The value `cmax` determines the maximum value `c` can be in order to guarantee `c`
                will be an unbiased estimate. In practice, it is often better to not modify this value and instead
                change `qpct` to control the biasedness of the estimate.
            cinit: The initial value to use for `c` (the rejection sampling multiplier). If left as None then a very
                conservative, data-adaptive estimate is used to initialize `c`. Without prior knowledge of the data
                leaving this as `None` is likely the best course of action.
            seed: Provide an explicit seed to use during evaluation. If not provided a default is used.
        """

        self._record = [record] if isinstance(record,str) else record
        self._ope    = ope
        self._cpct   = cpct
        self._cmax   = cmax
        self._cinit  = cinit
        self._seed   = seed

    @property
    def params(self) -> Mapping[str,Any]:
        return {'ope': self._ope, 'cpct': self._cpct, 'cmax': self._cmax, 'cinit': self._cinit, 'seed': self._seed }

    def evaluate(self, environment: Optional[Environment], learner: Optional[Learner]) -> Iterable[Mapping[Any,Any]]:

        #import here to avoid circular dependencies...
        from coba.environments import OpeRewards, Finalize, BatchSafe

        interactions = environment.read()
        learner      = SafeLearner(learner, self._seed if self._seed is not None else CobaContext.store.get("experiment_seed"))
        rng          = CobaRandom(self._seed if self._seed is not None else CobaContext.store.get("experiment_seed"))

        first_100, interactions = peek_first(interactions,n=100)

        if not interactions: return []

        first    = first_100[0]
        batched  = first and (is_batch(first.get('context')) or is_batch(first.get('actions')))
        discrete = 'actions' in first and len(first['actions'][0] if batched else first['actions']) > 0

        if not all(k in first.keys() for k in ['context', 'action', 'reward', 'actions', 'probability']):
            raise CobaException("ExplorationEvaluator requires interactions with `['context', 'action', 'reward', 'actions', 'probability']`")

        if not discrete:
            raise CobaException("ExplorationEvaluator does not currently support continuous actions")

        if batched:
            raise CobaException("ExplorationEvaluator does not currently support batching")

        if not learner.has_score:
            raise CobaException("ExplorationEvaluator requires Learners to implement a `score` method")

        record_time     = 'time'        in self._record
        record_reward   = 'reward'      in self._record
        record_action   = 'action'      in self._record
        record_actions  = 'actions'     in self._record
        record_context  = 'context'     in self._record
        record_prob     = 'probability' in self._record
        record_ope_loss = 'ope_loss'    in self._record

        score = learner.score
        learn = learner.learn
        pred  = learner.predict
        info  = CobaContext.learning_info

        info.clear()

        first_probs = [i['probability'] for i in first_100] + [(1-i['probability'])/(len(i['actions'])-1) for i in first_100]
        ope_rewards = []
        Q           = []
        c           = self._cinit or min(list(filter(None,first_probs))+[self._cmax])
        t           = 0

        ope_type = self._ope.upper() if self._ope else None

        interactions = BatchSafe(Finalize()).filter(interactions)

        if ope_type:
            interactions = BatchSafe(OpeRewards(ope_type,features=[1,'a','xa'])).filter(interactions)

        ope_ips = ope_type == 'IPS'

        for interaction in interactions:

            t += 1

            info.clear()
            interaction = interaction.copy()

            log_context      = interaction.pop('context')
            log_actions      = interaction.pop('actions')
            log_action       = interaction.pop('action')
            log_reward       = interaction.pop('reward')
            log_prob         = interaction.pop('probability')
            log_rewards      = interaction.pop('rewards',None)

            start_time = time.time()
            on_prob  = score(log_context,log_actions,log_action)
            predict_time = time.time()-start_time

            if ope_type:
                ope_rewards.append(on_prob*log_rewards(log_action) if ope_ips else log_rewards(pred(log_context,log_actions)[0]))

            #I tested many techniques for both maintaining Q and estimating its qpct percentile...
            #Implemented here is the insort method because it provided the best runtime by far.
            #The danger of this method is that the computational complexity is T (due to inserts).
            #Even so, T has to become very large (many millions?) before it is slower than alternatives.
            #The most obvious alternative I played with was using dequeue with a fixed size/complexity.
            #However, this requires sorting the dequeue for every accepted action which is incredibly slow.
            if on_prob != 0:
                insort(Q,log_prob/on_prob)

            #c \in log_prob/on_prob
            #so c is small when log_prob is small and on_prob is large

            #we want c*on_prob/log_prob <= 1 approximately (1 - qpct) of the time.
            #We know that if c <= min(log_prob) this condition will be met 100% of the time.
            #This might be too conservative though because it doesn't consider
            #the value of on_prob. For example if on_prob:=log_prob then the
            #above condition will be met if c = 1 which will be >= min(log_prob).
            if rng.random() <= c*(on_prob/log_prob):

                out = {}
                if ope_rewards:
                    ope_rewards[-1] = log_reward
                else:
                    ope_rewards.append(log_reward)

                start_time = time.time()
                learn(log_context, log_action, log_reward, on_prob)
                learn_time = time.time() - start_time

                if record_time    : out['predict_time'] = predict_time
                if record_time    : out['learn_time']   = learn_time
                if record_context : out['context']      = log_context
                if record_actions : out['actions']      = log_actions
                if record_action  : out['action']       = log_action
                if record_reward  : out['reward']       = mean(ope_rewards)
                if record_prob    : out['probability']  = on_prob
                if record_ope_loss: out['ope_loss']     = get_ope_loss(learner)

                if info: out.update(info)
                if out : yield out

                ope_rewards.clear()
                c = min(percentile(Q,self._cpct,sort=False), self._cmax)

        if ope_rewards:
            pass
            #If we hit this it means that there was rejected data at the end of the
            #simulation. I haven't found a good way of dealing with this case. In
            #general I don't think it should be a problem unless we are working with
            #very small datasets. Commented out below is the previous functionality:
            #out = {}
            #if record_time   : out['predict_time'] = predict_time
            #if record_reward : out['reward']       = mean(ope_rewards)
            #if out: yield out



from math import exp,dist
from operator import mul,add
from statistics import mean
from itertools import count, islice, repeat
from typing import Sequence, Tuple, Callable, Optional, Iterable, Literal, Mapping, Any, overload

from coba.random import CobaRandom
from coba.exceptions import CobaException
from coba.primitives import Context, Action
from coba.primitives import Environment, SimulatedInteraction
from coba.encodings import InteractionsEncoder, OneHotEncoder

class LambdaSimulation(Environment):
    """A contextual bandit environment created from generative lambda functions."""

    @overload
    def __init__(self,
        n_interactions: Optional[int],
        context       : Callable[[int               ],Context         ],
        actions       : Callable[[int,Context       ],Sequence[Action]],
        reward        : Callable[[int,Context,Action],float           ]) -> None:
        """Instantiate a LambdaSimulation.

        Args:
            n_interactions: An optional integer indicating the number of interactions in the simulation.
            context: A function that should return a context given an index in `range(n_interactions)`.
            actions: A function that should return all valid actions for a given index and context.
            reward: A function that should return the reward for the index, context and action.
        """

    @overload
    def __init__(self,
        n_interactions: Optional[int],
        context       : Callable[[int               ,CobaRandom],Context         ],
        actions       : Callable[[int,Context       ,CobaRandom],Sequence[Action]],
        reward        : Callable[[int,Context,Action,CobaRandom],float           ],
        seed          : int) -> None:
        """Instantiate a LambdaSimulation.

        Args:
            n_interactions: An optional integer indicating the number of interactions in the simulation.
            context: A function that should return a context given an index and random state.
            actions: A function that should return all valid actions for a given index, context and random state.
            reward: A function that should return the reward for the index, context, action and random state.
            seed: An integer used to seed the random state in order to guarantee repeatability.
        """

    def __init__(self,n_interactions,context,actions,reward,seed=None) -> None:
        """Instantiate a LambdaSimulation."""

        self._n_interactions = n_interactions
        self._context        = context
        self._actions        = actions
        self._reward         = reward
        self._make_rng       = seed is not None
        if seed is not None: self._seed = seed

    @property
    def params(self) -> Mapping[str, Any]:
        params = { "env_type": "LambdaSimulation" }

        if hasattr(self, '_seed'):
            params["seed"] = self._seed

        return params

    def read(self) -> Iterable[SimulatedInteraction]:
        rng = None if not self._make_rng else CobaRandom(self._seed)

        _context = lambda i    : self._context(i    ,rng) if rng else self._context(i   )
        _actions = lambda i,c  : self._actions(i,c  ,rng) if rng else self._actions(i,c )
        _reward  = lambda i,c,a: self._reward (i,c,a,rng) if rng else self._reward (i,c,a)

        for i in islice(count(), self._n_interactions):
            context  = _context(i)
            actions  = _actions(i, context)
            rewards  = [ _reward(i, context, action) for action in actions]

            yield {'context':context,'actions':actions,'rewards':rewards }

    def __str__(self) -> str:
        return "LambdaSimulation"

    class Spoof:
        def __init__(self, interactions, _params, _str, _name) -> None:
            self._str           = _str
            self._interactions  = interactions
            self._params        = _params
            type(self).__name__ = _name

        @property
        def params(self) -> Mapping[str, Any]:
            return self._params

        def read(self):
            return self._interactions

        def __str__(self) -> str:
            return self._str

    def __reduce__(self) -> Tuple[object, ...]:
        if self._n_interactions is not None and self._n_interactions < 5000:
            #This is an interesting idea but maybe too wink-wink nudge-nudge in practice. It causes a weird flow
            #in the logs that look like bugs. It also causes unexpected lags because IO is happening at strange
            #places and in a manner that can cause thread locks.
            return (LambdaSimulation.Spoof, (list(self.read()), self.params, str(self), type(self).__name__ ))
        else:
            message = (
                "It is not possible to pickle a LambdaSimulation due to its use of lambda methods in the constructor. "
                "This error occured because an experiment containing a LambdaSimulation tried to execute on multiple processes. "
                "If this is neccesary there are three options to get around this limitation: (1) run your experiment "
                "on a single process rather than multiple, (2) re-design your LambdaSimulation as a class that inherits "
                "from LambdaSimulation and implements __reduce__ (see coba.environments.simulations.LinearSyntheticSimulation "
                "for an example), or (3) specify a finite number for n_interactions in the LambdaSimulation constructor (this "
                "allows us to create the interactions in memory ahead of time and convert to an in-memory simulation to pickle).")
            raise CobaException(message)

class BanditSyntheticSimulation(Environment):
    """A simulation with fixed reward values for each action."""

    def __init__(self, n_interactions: Optional[int], n_actions: int, seed: Optional[int]=None) -> None:
        """Instantiate a BanditSimulation.

        Args:
            n_interactions: An optional integer indicating the number of interactions in the simulation.
            n_actions: The number of actions each interaction should have.
            seed: An integer used to seed the random state in order to guarantee repeatability.
        """

        self._n_interactions = n_interactions
        self._n_actions      = n_actions
        self._seed           = seed

    @property
    def params(self) -> Mapping[str, Any]:
        return { "env_type": "BanditSimulation", 'n_actions': self._n_actions, "seed": self._seed }

    def read(self) -> Iterable[SimulatedInteraction]:
        rewards = CobaRandom(self._seed).randoms(self._n_actions)
        actions = OneHotEncoder().fit_encodes(range(self._n_actions))

        yield from repeat({'context': None, 'actions': actions, 'rewards': rewards}, self._n_interactions)

    def __str__(self) -> str:
        return f"BanditSimulation(A={self._n_actions},seed={self._seed})"

class LinearSyntheticSimulation(Environment):
    """A synthetic simulation whose rewards are linear with respect to the given reward features.

    The simulation's rewards are linear with respect to the requrested reward features. When no context
    or action features are requested these terms are removed from the requested reward features.
    """

    def __init__(self,
        n_interactions:int,
        n_actions:int = 10,
        n_context_features:int = 10,
        n_action_features:int = 10,
        n_coefficients:Optional[int] = 5,
        reward_features:Sequence[str] = ["a","xa"],
        seed:int = 1) -> None:
        """Instantiate a LinearSyntheticSimulation.

        Args:
            n_interactions: The number of interactions the simulation should have.
            n_actions: The number of actions each interaction should have.
            n_context_features: The number of features each context should have.
            n_action_features: The number of features each action should have.
            n_coefficients The number of non-zero weights in the final reward function.
            reward_features: The features in the simulation's linear reward function.
            seed: The random number seed used to generate all features, weights and noise in the simulation.
        """

        if n_actions < 2:
            raise CobaException("Linear synthetic environments must have at least two actions")

        if isinstance(reward_features,str):
            reward_features = [reward_features]

        if n_action_features or n_context_features:
            feats = ''.join(reward_features)
            if not n_action_features: feats = feats.replace('a','')
            if not n_context_features: feats = feats.replace('x','')
            if not feats and n_context_features: raise CobaException("Reward features must include `x`")
            if not feats and n_action_features: raise CobaException("Reward features must include `a`")

        self._n_interactions     = n_interactions
        self._n_actions          = n_actions
        self._n_context_features = n_context_features
        self._n_action_features  = n_action_features
        self._n_coefficients     = n_coefficients
        self._reward_features    = reward_features
        self._seed               = seed

    def read(self) -> Iterable[SimulatedInteraction]:
        n_actions          = self._n_actions
        n_action_features  = self._n_action_features
        n_context_features = self._n_context_features
        n_coefficients     = self._n_coefficients
        reward_features    = self._reward_features

        if not n_context_features and not n_action_features:
            yield from BanditSyntheticSimulation(self._n_interactions, self._n_actions, self._seed).read()
            return

        replace = 'x' if not n_context_features else 'a' if not n_action_features else ''
        reward_features = list(filter(None,[f.replace(replace,'') for f in reward_features]))
        reward_features = sorted(set(reward_features),key=reward_features.index)

        rng           = CobaRandom(self._seed)
        feats_encoder = InteractionsEncoder(reward_features)
        feature_count = len(feats_encoder.encode(x=[1]*n_context_features,a=[1]*n_action_features))

        if n_action_features:
            output_size = 1 #f(x+a1) = r1; f(x+a2) = r2
        else:
            output_size = n_actions #f(x or 1) = [r1,r2,...]

        weights_count  = n_coefficients or feature_count or 1
        output_weights = [[0]*(feature_count or 1) for _ in range(output_size)]
        for i in range(output_size):
            indexes = rng.shuffle(range(feature_count or 1))
            weights = rng.randoms(weights_count,-1,1) if reward_features else rng.randoms(weights_count)
            for j,w in zip(indexes,weights):
                output_weights[i][j] = w

        output_biases  = [0] * output_size
        output_scalars = [1] * output_size

        phis    = lambda n: rng.randoms(n,-1,1)
        onehots = OneHotEncoder().fit_encodes(range(n_actions))

        calln       = (lambda callable,n: [callable() for _ in range(n)])
        context_gen = (lambda: phis(n_context_features   )) if n_context_features else (lambda: None   )
        action_gen  = (lambda: phis(n_action_features    )) if n_action_features  else (lambda: None   )
        actions_gen = (lambda: calln(action_gen,n_actions)) if n_action_features  else (lambda: onehots)
        feats_gen   = (lambda: feats_encoder.encode(x=next(context_iter),a=next(action_iter)))

        context_iter = iter(context_gen,'forever')
        action_iter  = iter(action_gen ,'forever')
        actions_iter = iter(actions_gen,'forever')
        feats_iter   = iter(feats_gen  ,'forever')

        def f(x):
            return [ bias + scalar * sum(map(mul,(x or [1]),weights)) for weights,bias,scalar in zip(output_weights,output_biases,output_scalars) ]

        get_range    = lambda r: max(r)-min(r)
        output_cols  = list(zip(*map(f,islice(feats_iter,100))))
        global_range = get_range(sum(output_cols,()))

        has_x = n_context_features >=1
        gt_1  = len(output_weights[0]) > 1

        for i,col in enumerate(output_cols):

            # when the reward functions are linear with respect to an individual
            # feature local scaling will make reward functions more or less identical
            output_range = get_range(col) if gt_1 else global_range if has_x else 1

            output_scalars[i] = 1/output_range            #scale to ~[0,1]
            output_biases[i]  = .5-mean(col)/output_range #center at ~.5

        for _ in range(self._n_interactions):
            context = next(context_iter)
            actions = next(actions_iter)

            if n_action_features:
                rewards = [f(feats_encoder.encode(x=context,a=action))[0] for action in actions]
            else:
                rewards = f(feats_encoder.encode(x=context))

            yield {'context': context, 'actions': actions, 'rewards': rewards}

    @property
    def params(self) -> Mapping[str, Any]:
        return {"env_type":"LinearSynthetic", "reward_features": self._reward_features, 'n_coeff':self._n_coefficients, 'n_actions': self._n_actions, "seed": self._seed}

    def __str__(self) -> str:
        return f"LinearSynth(A={self._n_actions},c={self._n_context_features},a={self._n_action_features},R={self._reward_features},seed={self._seed})"

class NeighborsSyntheticSimulation(Environment):
    """A synthetic simulation whose reward values are determined by neighborhoods.

    The simulation's rewards are determined by the location of given context and action pairs. These locations
    indicate which neighborhood the context action pair belongs to. Neighborhood rewards are determined by
    random assignment.
    """

    def __init__(self,
        n_interactions:int,
        n_actions:int = 10,
        n_context_features:int = 10,
        n_action_features:int = 10,
        n_neighborhoods:int = 10,
        seed: int = 1) -> None:
        """Instantiate a NeighborsSyntheticSimulation.

        Args:
            n_interactions: The number of interactions the simulation should have.
            n_actions: The number of actions each interaction should have.
            n_context_features: The number of features each context should have.
            n_action_features: The number of features each action should have.
            n_neighborhoods: The number of neighborhoods the simulation should have.
            seed: The random number seed used to generate all contexts and action rewards.
        """

        self._n_interactions  = n_interactions
        self._n_actions       = n_actions
        self._n_context_feats = n_context_features
        self._n_action_feats  = n_action_features
        self._n_neighborhoods = n_neighborhoods
        self._seed            = seed

    def read(self) -> Iterable[SimulatedInteraction]:
        n_interactions  = self._n_interactions
        n_actions       = self._n_actions
        n_context_feats = self._n_context_feats
        n_action_feats  = self._n_action_feats
        n_neighborhoods = self._n_neighborhoods

        if not n_context_feats and not n_action_feats:
            yield from BanditSyntheticSimulation(self._n_interactions, self._n_actions, self._seed).read()
            return

        rng = CobaRandom(self._seed)

        phis    = lambda n: rng.randoms(n,0,1)
        onehots = OneHotEncoder().fit_encodes(range(n_actions))

        calln       = (lambda callable,n: [callable() for _ in range(n)])
        context_gen = (lambda: phis(n_context_feats      )) if n_context_feats else (lambda: None   )
        action_gen  = (lambda: phis(n_action_feats       )) if n_action_feats  else (lambda: None   )
        actions_gen = (lambda: calln(action_gen,n_actions)) if n_action_feats  else (lambda: onehots)

        context_iter = iter(context_gen,'forever')
        action_iter  = iter(action_gen ,'forever')
        actions_iter = iter(actions_gen,'forever')

        if n_action_feats == 0:
            worlds = [list(zip(context_iter,rng.randoms(n_neighborhoods))) for _ in range(n_actions)]
        else:
            worlds = [list(zip(map(add,(c or [] for c in context_iter),action_iter),rng.randoms(n_neighborhoods))) for _ in range(n_actions)]

        def f(x):
            return [ min(world, key=lambda w: dist(w[0],x))[1] for world in worlds ]

        self.worlds = worlds

        for _ in range(n_interactions):

            context = next(context_iter)
            actions = next(actions_iter)

            if n_action_feats:
                rewards = [ f((context or [])+action)[0] for action in actions]
            else:
                rewards = f(context)

            yield {'context': context, 'actions': actions, 'rewards': rewards}

    @property
    def params(self) -> Mapping[str, Any]:
        return {"env_type": "NeighborsSynthetic", "n_neighborhoods": self._n_neighborhoods, 'n_actions': self._n_actions, 'seed': self._seed }

    def __str__(self) -> str:
        return f"NeighborsSynth(A={self._n_actions},c={self._n_context_feats},a={self._n_action_feats},N={self._n_neighborhoods},seed={self._seed})"

class KernelSyntheticSimulation(Environment):
    """A synthetic simulation whose reward function is created from kernel basis functions.

    Kernel functions are created using random exemplar points generated at initialization and fixed for all time.
    """

    def __init__(self,
        n_interactions:int,
        n_actions:int = 10,
        n_context_features:int = 10,
        n_action_features:int = 10,
        n_exemplars:int = 10,
        kernel: Literal['linear','polynomial','exponential','gaussian'] = 'gaussian',
        degree: int = 2,
        gamma: float = 1,
        seed: int = 1) -> None:
        """Instantiate a KernelSyntheticSimulation.

        Args:
            n_interactions: The number of interactions the simulation should have.
            n_actions: The number of actions each interaction should have.
            n_context_features: The number of features each context should have.
            n_action_features: The number of features each action should have.
            n_exemplars: The number of exemplar context-action pairs.
            kernel: The family of the kernel basis functions.
            degree: This argument is only relevant when using polynomial kernels.
            gamma: This argument is only relevant when using exponential kernels.
            seed: The random number seed used to generate all features, weights and noise in the simulation.
        """

        if n_actions < 2:
            raise CobaException("Kernel synthetic environments must have at least two actions.")

        if n_exemplars < 1:
            raise CobaException("Kernel synthetic environments must have at least one exemplar.")

        self._n_interactions     = n_interactions
        self._n_actions          = n_actions
        self._n_context_features = n_context_features
        self._n_action_features  = n_action_features
        self._n_exemplars        = n_exemplars
        self._seed               = seed
        self._kernel             = kernel
        self._degree             = degree
        self._gamma              = gamma

    def read(self) -> Iterable[SimulatedInteraction]:

        rng = CobaRandom(self._seed)

        n_interactions     = self._n_interactions
        n_actions          = self._n_actions
        n_context_features = self._n_context_features
        n_action_features  = self._n_action_features
        n_exemplars        = self._n_exemplars
        kernel             = self._kernel

        if not n_context_features and not n_action_features:
            yield from BanditSyntheticSimulation(self._n_interactions, self._n_actions, self._seed).read()
            return

        if kernel == "linear":
            K = lambda x1,x2: self._linear_kernel(x1,x2)
        if kernel == "polynomial":
            K = lambda x1,x2: self._polynomial_kernel(x1,x2,self._degree)
        if kernel == "exponential":
            K = lambda x1,x2: self._exponential_kernel(x1,x2,self._gamma)
        if kernel == "gaussian":
            K = lambda x1,x2: self._gaussian_kernel(x1,x2,self._gamma)

        if n_action_features:
            output_size = 1 #f(x+a1) = r1; f(x+a2) = r2
        else:
            output_size = n_actions #f(x) = [r1,r2,...]

        n_exemplar_features = n_action_features+n_context_features

        phis    = lambda n: rng.randoms(n,0,1)
        onehots = OneHotEncoder().fit_encodes(range(n_actions))

        calln        = (lambda callable,n: [callable() for _ in range(n)])
        context_gen  = (lambda: phis(n_context_features   )) if n_context_features else (lambda: None   )
        action_gen   = (lambda: phis(n_action_features    )) if n_action_features  else (lambda: None   )
        actions_gen  = (lambda: calln(action_gen,n_actions)) if n_action_features  else (lambda: onehots)
        exemplar_gen = (lambda: phis(n_exemplar_features  ))

        context_iter  = iter(context_gen ,'forever')
        actions_iter  = iter(actions_gen ,'forever')
        exemplar_iter = iter(exemplar_gen,'forever')

        output_exemplars = [ calln(exemplar_gen, n_exemplars) for _ in range(output_size) ]
        output_weights   = [ rng.randoms(n_exemplars, -1, 1)   for _ in range(output_size) ]
        output_biases    = [0] * output_size
        output_scalars   = [1] * output_size

        def f(x):
            rewards=[]
            for exemplars,weights,bias,scalar in zip(output_exemplars,output_weights,output_biases,output_scalars):
                reward = 0
                for w,e in zip(weights,exemplars):
                    reward += w*K(e,x)
                rewards.append(bias + scalar*reward)
            return rewards

        get_range    = lambda r: max(r)-min(r)
        output_cols  = list(zip(*map(f,islice(exemplar_iter,100))))
        global_range = get_range(sum(output_cols,()))

        gt_1 = n_exemplar_features > 1
        linear = kernel == 'linear' or (kernel == 'polynomial' and self._degree == 1)

        for i,col in enumerate(output_cols):

            # when the reward functions are linear with respect to an individual
            # feature local scaling will make reward functions more or less identical
            output_range = get_range(col) if gt_1 or not linear else global_range

            output_scalars[i] = 1/output_range            #scale to ~[0,1]
            output_biases[i]  = .5-mean(col)/output_range #center at ~.5

        for _ in range(n_interactions):

            context = next(context_iter)
            actions = next(actions_iter)

            if n_action_features:
                rewards = [f((context or [])+action)[0] for action in actions]
            else:
                rewards = f(context)

            yield {'context': context, 'actions': actions, 'rewards': rewards}

    @property
    def params(self) -> Mapping[str, Any]:
        params = {"env_type": "KernelSynthetic", "n_exemplars": self._n_exemplars, 'kernel': self._kernel}

        if self._kernel == "polynomial":
            params['degree'] = self._degree

        if self._kernel in ["exponential","gaussian"]:
            params['gamma'] = self._gamma

        params['n_actions'] = self._n_actions
        params['seed'] = self._seed

        return params

    def __str__(self) -> str:
        return f"KernelSynth(A={self._n_actions},c={self._n_context_features},a={self._n_action_features},E={self._n_exemplars},K={self._kernel},seed={self._seed})"

    def _linear_kernel(self, F1: Sequence[float], F2: Sequence[float]) -> float:
        return sum([f1*f2 for f1,f2 in zip(F1,F2)])

    def _polynomial_kernel(self, F1: Sequence[float], F2: Sequence[float], degree:int) -> float:
        return (self._linear_kernel(F1,F2)+1)**degree

    def _exponential_kernel(self, F1: Sequence[float], F2: Sequence[float], gamma:float) -> float:
        return exp(-dist(F1,F2)/gamma)

    def _gaussian_kernel(self, F1: Sequence[float], F2: Sequence[float], gamma:float) -> float:
        return exp(-dist(F1,F2)**2/gamma)

class MLPSyntheticSimulation(Environment):
    """A synthetic simulation whose reward function belongs to the MLP family.

    The MLP architecture has a single hidden layer with sigmoid activation and one output
    value calculated from a random linear combination of the hidden layer's output.
    """

    def __init__(self,
        n_interactions:int,
        n_actions:int = 10,
        n_context_features:int = 10,
        n_action_features:int = 10,
        seed: int = 1) -> None:
        """Instantiate an MLPSythenticSimulation.

        Args:
            n_interactions: The number of interactions the simulation should have.
            n_actions: The number of actions each interaction should have.
            n_context_features: The number of features each context should have.
            n_action_features: The number of features each action should have.
            seed: The random number seed used to generate all features, weights and noise in the simulation.
        """

        if n_actions < 2:
            raise CobaException("MLP synthetic environments must have at least two actions.")

        self._n_actions          = n_actions
        self._n_context_features = n_context_features
        self._n_action_features  = n_action_features
        self._n_interactions     = n_interactions
        self._seed               = seed

    def read(self) -> Iterable[SimulatedInteraction]:

        rng = CobaRandom(self._seed)

        n_context_features = self._n_context_features
        n_action_features  = self._n_action_features
        n_actions          = self._n_actions

        if not n_context_features and not n_action_features:
            yield from BanditSyntheticSimulation(self._n_interactions, self._n_actions, self._seed).read()
            return

        # a lot of emprical experiments showed
        # these values hit a sweet spot in terms
        # of reward variation and complexity
        hidden = 20
        power  = 32

        if n_action_features:
            #f(x+a1) = r1; f(x+a2) = r2
            input_size  = n_context_features + n_action_features
            hidden_size = hidden
            output_size = 1
        else:
            #f(x or 1) = [r1,r2,...]
            input_size  = n_context_features or 1
            hidden_size = hidden
            output_size = n_actions

        hidden_weights    = [ rng.gausses(input_size,0,1.5) for _ in range(hidden_size) ]
        hidden_activation = lambda x: 1/(1+exp(-x)) #sigmoid activation
        output_weights    = [ [ w**power for w in rng.randoms(hidden_size,0,1)] for _ in range(output_size) ]
        output_weights    = [ [ w/sum(weights) for w in weights ] for weights in output_weights ]

        if n_context_features:
            context_iter = iter(rng.gausses(n_context_features) for _ in repeat(1))
        else:
            context_iter = iter(repeat(None))

        if n_action_features:
            actions_iter = iter([rng.gausses(n_action_features) for _ in range(n_actions)] for _ in repeat(1))
        else:
            actions_iter = iter(repeat(OneHotEncoder().fit_encodes(range(n_actions))))

        def f(input):
            hidden_out = [ hidden_activation(sum(map(mul,input,weights))) for weights in hidden_weights ]
            output_val = [ sum(map(mul,hidden_out,weights))               for weights in output_weights ]
            return output_val

        for _ in range(self._n_interactions):

            context = next(context_iter)
            actions = next(actions_iter)

            if n_action_features:
                rewards = [ f( (context or []) + action )[0] for action in actions ]
            else:
                rewards = f(context or [1])

            yield {'context': context, 'actions': actions, 'rewards': rewards}

    @property
    def params(self) -> Mapping[str, Any]:
        return {"env_type": "MLPSynthetic", 'n_actions': self._n_actions, 'seed': self._seed}

    def __str__(self) -> str:
        return f"MLPSynth(A={self._n_actions},c={self._n_context_features},a={self._n_action_features},seed={self._seed})"



import math

from collections import defaultdict, abc
from itertools import islice, chain
from typing import Iterable, Any, Sequence, Mapping, Optional, Union, Iterator

from coba.random import CobaRandom
from coba.encodings import Encoder
from coba.utilities import peek_first, try_else
from coba.primitives import Sparse, Dense, Filter
from coba.pipes.utilities import resolve_params

class FiltersFilter(Filter):
    def __init__(self, *pipes: Filter):
        self._filters = sum((try_else(lambda: list(p),[p]) for p in pipes),[])

    @property
    def params(self) -> Mapping[str,Any]:
        return resolve_params(list(self))

    def filter(self, items: Any) -> Any:
        for filter in self._filters:
            items = filter.filter(items)
        return items

    def __str__(self) -> str:
        return " | ".join(map(str,self._filters))

    def __getitem__(self, index: int) -> Filter:
        return self._filters[index]

    def __iter__(self) -> Iterator[Filter]:
        return iter(self._filters)

    def __len__(self) -> int:
        return len(self._filters)

class Identity(Filter[Any, Any]):
    """A filter which returns what is given."""
    def filter(self, item:Any) -> Any:
        return item

    @property
    def params(self) -> Mapping[str, Any]:
        return {}

class Shuffle(Filter[Iterable[Any], Sequence[Any]]):
    """Shuffle a sequence of items."""

    def __init__(self, seed:Optional[int]) -> None:
        """Instantiate a Shuffle filter.

        Args:
            seed: A random number seed which determines the new sequence order.
        """

        if seed is not None and (not isinstance(seed,int) or seed < 0):
            raise ValueError(f"Invalid parameter for Shuffle: {seed}. An optional integer value >= 0 was expected.")

        self._seed = seed

    def filter(self, items: Iterable[Any]) -> Sequence[Any]:
        items = items.copy() if isinstance(items,list) else list(items)
        yield from CobaRandom(self._seed).shuffle(items,inplace=True)

    @property
    def params(self) -> Mapping[str, Any]:
        return { "shuffle_seed": self._seed }

class Take(Filter[Iterable[Any], Sequence[Any]]):
    """Take a fixed number of items from an iterable."""

    def __init__(self, count: Optional[int], strict:bool = False) -> None:
        """Instantiate a Take filter.

        Args:
            count: The number of items we wish to take from an iterable.
            strict: Whether we want to take anything up to count or only count exactly.
        """

        is_valid_count = (count is None) or (isinstance(count,int) and count >= 0)

        if not is_valid_count:
            raise ValueError(f"Invalid value for Take: {count}. A positive integer or None was expected.")

        self._count = count
        self._strict = strict

    def filter(self, items: Iterable[Any]) -> Iterable[Any]:
        out = islice(items, self._count)
        if self._strict: out = list(out)
        return [] if self._strict and len(out) < self._count else out

    @property
    def params(self) -> Mapping[str, Any]:
        return { "take": self._count }

class Slice(Filter[Iterable[Any], Sequence[Any]]):
    """Take a slice of items from an iterable."""

    def __init__(self, start:Optional[int], stop: Optional[int], step:int=1) -> None:
        """Instantiate a Slice filter.

        Args:
            start: The index where the slice starts.
            stop: The index where the slice stops.
            step: The step size between each item in the slice.
        """

        is_valid_start = (start is None) or (isinstance(start,int) and start >= 0)
        is_valid_stop  = (stop is None) or (isinstance(stop,int) and stop >= 0)

        if not is_valid_start or not is_valid_stop:
            raise ValueError(f"Invalid value for start or stop was given to Slice.")

        self._start = start
        self._stop = stop
        self._step = step

    def filter(self, items: Iterable[Any]) -> Iterable[Any]:
        return islice(items, self._start, self._stop, self._step)

    @property
    def params(self) -> Mapping[str, Any]:
        params = { "slice_start": self._start, "slice_stop": self._stop }
        if self._step != 1: params['slice_step'] = self._step
        return params

class Reservoir(Filter[Iterable[Any], Sequence[Any]]):
    """Take a fixed number of random items from an iterable.

    Remarks:
        We use Algorithm L as described by Kim-Hung Li. (1994) to take a random count of items.

    References:
        Kim-Hung Li. 1994. Reservoir-sampling algorithms of time complexity O(n(1 + log(N/n))).
        ACM Trans. Math. Softw. 20, 4 (Dec. 1994), 481–493. DOI:https://doi.org/10.1145/198429.198435
    """

    def __init__(self, count: Optional[int], strict: bool = False, seed: float = 1) -> None:
        """Instantiate a Reservoir filter.

        Args:
            count: The number of items we wish to sample from an iterable (expressed as either an exact value or range).
            strict: Whether we want to take anything up to count or only count exactly.
            seed : The seed which determines which random items to take.
        """

        is_valid_count = (count is None) or (isinstance(count,int) and count >= 0)

        if not is_valid_count:
            raise ValueError(f"Invalid value for Reservoir: {count}. A positive integer or None was expected.")

        self._count  = count
        self._seed   = seed
        self._strict = strict

    @property
    def params(self) -> Mapping[str, Any]:
        return { "reservoir_count": self._count, "reservoir_seed": self._seed }

    def filter(self, items: Iterable[Any]) -> Sequence[Any]:

        rng = CobaRandom(self._seed)

        if self._count == 0:
            yield from []

        elif self._count is None:
            yield from rng.shuffle(items)

        else:
            W         = 1
            items     = iter(items)
            reservoir = list(islice(items,self._count))

            if len(reservoir) < self._count:
                yield from ([] if self._strict else rng.shuffle(reservoir,inplace=True))

            else:
                reservoir = rng.shuffle(reservoir,inplace=True)
                count     = self._count or 1
                log       = math.log
                floor     = math.floor
                x         = 1/count

                def batched_randoms_forever(batch_size):
                    while True:
                        randoms = rng.randoms(3*batch_size)
                        for i in range(0,3*batch_size,3):
                            yield randoms[i:i+3]

                try:
                    for r1,r2,r3 in batched_randoms_forever(20):
                        W = W*r1**x
                        S = floor(log(r2,1-W))
                        reservoir[int(r3*count)] = next(islice(items,S,S+1))
                except StopIteration:
                    pass

                yield from reservoir

class Flatten(Filter[Iterable[Any], Iterable[Any]]):
    """A filter which flattens rows in table shaped data."""

    def filter(self, data: Iterable[Any]) -> Iterable[Any]:

        first, data = peek_first(data)

        if isinstance(first,abc.MutableSequence):
            first_type = "list"
        elif isinstance(first,Dense):
            first_type = "tuple"
        elif isinstance(first,Sparse):
            first_type = "dict"
        else:
            first_type = None

        is_flattable = lambda v: isinstance(v,Dense) or not isinstance(v,str) and (isinstance(v,abc.Iterable))

        if first_type in ["list","tuple"]:
            flattable = [ is_flattable(v) for v in first ]
            any_flattable = any(flattable)
        elif first_type == "dict":
            #we assume that flattable entries will be populated in every row even though the dataset is sparse
            flattable = {k:[f"{k}_{i}" for i in range(len(v))] for k,v in first.items() if is_flattable(v) }
            any_flattable = bool(flattable)
        else:
            flattable = []
            any_flattable = False

        def flatter_list(row):
            for f,r in zip(flattable,row):
                if f: yield from r
                else: yield r

        if not any_flattable or first_type is None:
            yield from data
        elif first_type == "list":
            yield from (list(flatter_list(row)) for row in data)
        elif first_type == "tuple":
            yield from (tuple(flatter_list(row)) for row in data)
        elif first_type == "dict":
            for row in data:
                #in-line dict comprehension was faster than a generator like we did with list
                yield {k:v for k,v in row.items() for k,v in (zip(flattable[k],v) if k in flattable else ((k,v),)) if v != 0 }

class Encode(Filter[Iterable[Union[Sequence,Mapping]], Iterable[Union[Sequence,Mapping]]]):
    """A filter which encodes features in table shaped data."""

    def __init__(self, encoders:Mapping[Any,Encoder], fit_using:int = None, missing_val: str = None):
        """Instantiate an Encode filter.

        Args:
            encoders: A mapping from feature names to the Encoder to be used on the data.
            fit_using: The number of rows that should be used to fit the encoder.
            missing_val: The value that should be used when a feature is missing from a row.
        """
        self._encoders    = encoders
        self._fit_using   = fit_using
        self._missing_val = missing_val

    def filter(self, items: Iterable[Union[Sequence,Mapping]]) -> Iterable[Union[Sequence,Mapping]]:

        items = iter(items) # this makes sure items are pulled out for fitting

        try:
            first_item = next(items)
            items      = chain([first_item], items)
        except StopIteration:
            return []

        is_dense = isinstance(first_item,Dense)

        if not is_dense:
            encoders = self._encoders
        else:
            encoders = {}
            for k,v in self._encoders.items():
                try:
                    first_item[k]
                    encoders[k] = v
                except:
                    pass

        unfit_enc = { k:v for k,v in encoders.items() if not v.is_fit }
        fit_using = self._fit_using if unfit_enc else 0

        items_for_fitting  = list(islice(items, fit_using))
        values_for_fitting = defaultdict(list)

        for item in items_for_fitting:
            for k in unfit_enc:
                val = item[k] if is_dense else item.get(k,0)
                if val not in  ['',self._missing_val]: values_for_fitting[k].append(val)

        for k,v in values_for_fitting.items():
            encoders[k] = encoders[k].fit(v)

        for item in chain(items_for_fitting, items):
            item = list(item) if is_dense else dict(item)
            for k,v in encoders.items():
                if is_dense or k in item:
                    val = item[k]
                    item[k] = encoders[k].encode(val) if val not in ['',self._missing_val] else val
            yield item

class Structure(Filter[Iterable[Union[Sequence,Mapping]], Iterable[Any]]):
    """A filter which restructures rows in table shaped data."""

    def __init__(self, structure: Sequence[Any]) -> None:
        """Instantiate Structure filter.

        Args:
            structure: The structure that each row should be reformed into. Items in the structure should be
                feature names. A None value indicates the location that all left-over features will be placed.
        """

        self._structure = []
        stack = [structure]
        while stack:
            item = stack.pop()
            if isinstance(item,list):
                stack.append("LO")
                stack.extend(item)
                stack.append("LC")
            elif isinstance(item,tuple):
                stack.append("TO")
                stack.extend(item)
                stack.append("TC")
            else:
                self._structure.insert(0,item)

    def filter(self, data: Iterable[Union[Sequence,Mapping]]) -> Iterable[Any]:
        first, data = peek_first(data)
        is_dense = isinstance(first, Dense)

        for row in data:

            row = list(row) if is_dense else dict(row.items())

            stack   = []
            working = []
            for item in self._structure:
                if item in ["LO","TO"]:
                    stack.append(working)
                    working = []
                elif item == "LC":
                    new_working = stack.pop()
                    new_working.append(working)
                    working = new_working
                elif item == "TC":
                    new_working = stack.pop()
                    new_working.append(tuple(working))
                    working = new_working
                elif item == None:
                    working.append(row)
                else:
                    working.append(row.pop(item))

            yield working[0]

class Default(Filter[Iterable[Union[Sequence,Mapping]], Iterable[Union[Sequence,Mapping]]]):
    """A filter which sets default values for row features in table shaped data."""

    def __init__(self, defaults: Mapping[Any, Any]) -> None:
        """Instantiate a Default filter.

        Args:
            defaults: A mapping from feature names to their default values.
        """
        self._defaults = defaults

    def filter(self, data: Iterable[Union[Sequence,Mapping]]) -> Iterable[Union[Sequence,Mapping]]:

        for row in data:

            if isinstance(row, Sparse):
                row = dict(row)
                for k,v in self._defaults.items():
                    if k not in row: row[k] = v

            yield row

class Cache(Filter[Iterable[Any], Iterable[Any]]):

    def __init__(self,n_slice:int=25,protected:bool=False) -> None:
        self._cache     = None
        self._iter      = None
        self._n_slice   = n_slice
        self._protected = protected

    @property
    def protected(self) -> bool:
        return self._protected

    def filter(self, items: Iterable[Any]) -> Iterable[Any]:
        n_slice = self._n_slice

        if self._iter is None and self._cache is None:
            self._iter  = iter(items)
            self._cache = []

        if self._cache is not None and self._iter is None:
            yield from self._cache
            return

        yield from self._cache
        items = self._iter
        while current := list(islice(self._iter,n_slice)):
            self._cache.extend(current)
            yield from current
        self._iter = None

class Insert(Filter[Iterable[Any], Iterable[Any]]):
    def __init__(self, insert_items: Sequence[Any]) -> None:
        self._insert_items = insert_items

    def filter(self, items: Iterable[Any]) -> Iterable[Any]:
        yield from self._insert_items
        yield from items



from typing import Tuple, Callable,Sequence

from coba.random import CobaRandom
from coba.primitives import Context, Action, Actions, Prob, Kwargs

class PMFPredictor:
    def __init__(self, pmf: Callable[['Context','Actions'],Sequence[Prob]], seed:int = 1) -> None:
        """Instantiate a PMFPredictor.

        Args:
            pmf: Return the PMF for given context and actions.
            seed: A seed for a random number generation.
        """
        self._pmfrng  = CobaRandom(seed)
        self._pmfcall = pmf

    @property
    def seed(self) -> int:
        return self._pmfrng.seed

    def score(self, context: 'Context', actions: 'Actions', action: 'Action') -> 'Prob':
        return self._pmfcall(context,actions)[actions.index(action)]

    def predict(self, context: 'Context', actions: 'Actions') -> Tuple['Action','Prob']:
        return self._pmfrng.choicew(actions,self._pmfcall(context,actions))

class PMFInfoPredictor:
    def __init__(self, pmf: Callable[['Context','Actions'],Tuple[Sequence[Prob],'Kwargs']], seed:int = 1) -> None:
        """Instantiate a PMFInfoPredictor.

        Args:
            pmf: Return the PMF for given context and actions.
            seed: A seed for a random number generation.
        """
        self._pmfrng  = CobaRandom(seed)
        self._pmfcall = pmf

    @property
    def seed(self) -> int:
        return self._pmfrng.seed

    def score(self, context: 'Context', actions: 'Actions', action: 'Action') -> 'Prob':
        return self._pmfcall(context,actions)[0][actions.index(action)]

    def predict(self, context: 'Context', actions: 'Actions') -> Tuple['Action','Prob','Kwargs']:
        pmf,info = self._pmfcall(context,actions)
        return *self._pmfrng.choicew(actions,pmf),info



from coba.random import CobaRandom
from coba.context import CobaContext, Logger, NullLogger, BasicLogger, IndentLogger, NullCacher, DiskCacher

from coba.environments import Environments, ArffSource, CsvSource, LibSvmSource, ManikSource
from coba.environments import LambdaSimulation

from coba.learners import FixedLearner, RandomLearner
from coba.learners import BanditEpsilonLearner, BanditUCBLearner
from coba.learners import CorralLearner, LinUCBLearner, LinTSLearner
from coba.learners import VowpalLearner, VowpalEpsilonLearner, VowpalSoftmaxLearner, VowpalBagLearner, VowpalRndLearner
from coba.learners import VowpalCoverLearner, VowpalRegcbLearner, VowpalSquarecbLearner, VowpalOffPolicyLearner, VowpalMediator
from coba.learners import MisguidedLearner

from coba.evaluators import ClassMetaEvaluator, RejectionCB, SequentialCB, SequentialIGL
from coba.experiments import Experiment

from coba.results import Result, PointAndInterval, StdDevCI, StdErrCI, BootstrapCI, BinomialCI, Missing

from coba.encodings import InteractionsEncoder
from coba.utilities import peek_first, minimize
from coba.exceptions import CobaException
from coba.safety import SafeLearner
from coba.statistics import mean

from coba.primitives import is_batch, Context, Action, Actions, Categorical, Dense, Sparse
from coba.primitives import Learner, Environment, Interaction, Evaluator, Rewards, Namespaces
from coba.primitives import LoggedInteraction, SimulatedInteraction, GroundedInteraction
from coba.primitives import L1Reward, HammingReward, DiscreteReward

__version__ = "8.0.4"



import time
import json

from urllib import request
from operator import attrgetter
from typing import Tuple, Sequence, Any, Iterable, Dict, MutableSequence, MutableMapping, Union, overload

from coba.random import random
from coba.pipes import Pipes, HttpSource, ArffReader, DropRows, LabelRows
from coba.context import CobaContext
from coba.primitives import Sparse, Dense, Source
from coba.exceptions import CobaException

from coba.environments.supervised import SupervisedSimulation

class OpenmlSource(Source[Iterable[Tuple[Union[MutableSequence, MutableMapping],Any]]]):
    """Load an openml.org dataset."""

    @overload
    def __init__(self, *, data_id:int, drop_missing:bool=True, target:str = None):
        """Instantiate an OpenmlSource.
        Args:
            data_id: The data id uniquely identifying the dataset on openml (i.e., openml.org/d/{id})
            drop_missing: Drop data rows with missing values.
            target: The column that should be marked as the prediction label in the source.
        """
        ...

    @overload
    def __init__(self, *, task_id:int, drop_missing:bool=True, target:str = None):
        """Instantiate an OpenmlSource.
        Args:
            task_id: The openml task id which identifies the dataset to use from openml along with its label
            drop_missing: Drop data rows with missing values.
            target: The column that should be marked as the prediction label in the source
        """
        ...

    def __init__(self, **kwargs):
        """Instantiate an OpenmlSource."""

        self._data_id        = kwargs.get('data_id',None)
        self._task_id        = kwargs.get('task_id',None)
        self._target         = kwargs.get("target", None)
        self._drop_missing   = kwargs.get('drop_missing',True)

    @property
    def params(self) -> Dict[str,Any]:
        """Parameters describing the openml source."""
        return  { "openml_data": self._data_id, "openml_task": self._task_id, "openml_target": self._target, "drop_missing": self._drop_missing }

    def read(self) -> Iterable[Union[Dense,Sparse]]:
        """Read and parse the openml source."""

        try:

            # we only allow three paralellel request, an attempt at being "considerate" to openml
            # openml_semaphore will be None if we aren't multiprocessing otherwise it'll have a sema
            openml_semaphore = CobaContext.store.get("openml_semaphore")
            semaphore_acquired = False

            if openml_semaphore and not self._source_already_cached():
                openml_semaphore.acquire()
                if self._source_already_cached(): #pragma: no cover
                    openml_semaphore.release() #in-case another process cached everything needed while we were waiting
                else:
                    semaphore_acquired = True

            if self._data_id:
                data_descr   = self._get_data_descr(self._data_id)
                self._target = self._target or self._clean_name(data_descr.get("default_target_attribute",None))
                task_type    = None

            if self._task_id:
                task_descr = self._get_task_descr(self._task_id)
                task_type  = task_descr['type']

                if task_descr['type'] not in [1,2]:
                    raise CobaException(f"Openml task {self._task_id} does not appear to be a regression or classification task.")

                if not task_descr['data']:
                    raise CobaException(f"Openml task {self._task_id} does not appear to have an associated data source.")

                self._data_id = task_descr['data']
                self._target  = self._clean_name(task_descr.get('target'))
                data_descr    = self._get_data_descr(self._data_id)

            if not self._target:
                raise CobaException(f"We were unable to find an appropriate target column for the given openml source.")

            if data_descr.get('status') == 'deactivated':
                raise CobaException(f"Openml {self._data_id} has been deactivated. This is often due to flags on the data.")

            is_ignore = lambda feat_descr:(
                feat_descr['is_ignore'        ] == 'true' or
                feat_descr['is_row_identifier'] == 'true' or
                feat_descr['data_type'        ] not in ['numeric', 'nominal']
            )

            ignore = [ self._clean_name(f['name']) for f in self._get_feat_descr(self._data_id) if is_ignore(f)]

            if self._target in ignore: ignore.pop(ignore.index(self._target))

            label_type = 'c' if task_type==1 else 'r' if task_type==2 else None
            drop_row = attrgetter('missing') if self._drop_missing else None

            lines = self._get_arff_lines(data_descr["file_id"], None)
            reader= ArffReader()
            drop  = DropRows(drop_cols=ignore, drop_row=drop_row)
            label = LabelRows(self._target, label_type)

            yield from Pipes.join(reader, drop, label).filter(lines)

        except KeyboardInterrupt:
            #we don't want to clear the cache in the case of a KeyboardInterrupt
            raise

        except CobaException:
            #we don't want to clear the cache if it is an error we know about (the original raise should clear if needed)
            raise

        except Exception:
            #if something unexpected went wrong clear the cache just in case it was corrupted somehow
            self._clear_cache()
            raise

        finally:
            if semaphore_acquired:
                openml_semaphore.release()

    def _clean_name(self, name: str) -> str:
        return name.strip().strip('\'"').replace('\\','') if name else name

    def _get_data(self, url:str, key:str, checksum:str=None) -> Iterable[str]:
        try:
            with CobaContext.cacher.get_set(key, lambda: self._http_request(url)) as out:
                yield from out
        except Exception:
            self._clear_cache()
            raise

    def _http_request(self, url: str, tries: int = 0) -> Iterable[str]:
        api_key   = CobaContext.api_keys.get('openml')
        semaphore = CobaContext.store.get("openml_semaphore")

        # In an attempt to be considerate we stagger/limit our hits of the REST API.
        # Openml doesn't publish any rate-limiting guidelines, so this is just a guess.
        # if semaphore is not None it indictes that we are in a CobaMultiprocessor.
        if semaphore: time.sleep(2*random())

        try:
            KB = 1024
            MB = 1024*KB
            if api_key: url = f"{url}?api_key={api_key}"
            yield from HttpSource(url, timeout=20, chunk_size=10*MB).read()

        except TimeoutError:
            if tries >= 3: raise
            yield from self._http_request(url, tries+1)
            return

        except request.HTTPError as e:
            status, content = e.code, e.fp.read()

            if status == 412 and 'please provide api key' in content.lower():
                raise CobaException(
                    "Openml has requested an API key to access openml's rest API. A key can be obtained by creating "
                    "an openml account at openml.org. Once a key has been obtained it should be placed within "
                    "~/.coba as { \"api_keys\" : { \"openml\" : \"<your key here>\", } }.")

            if status == 412 and 'authentication failed' in content.lower():
                raise CobaException(
                    "The API key you provided no longer seems to be valid. You may need to create a new one by "
                    "logging into your openml account and regenerating a key. After regenerating the new key "
                    "should be placed in ~/.coba as { \"api_keys\" : { \"openml\" : \"<your key here>\", } }.")

            if status == 404:
                raise CobaException("We're sorry but we were unable to find the requested dataset on openml.")

            raise CobaException(f"An error was returned by openml: {content}")


    def _get_data_descr(self, data_id:int) -> Dict[str,Any]:
        descr_txt = " ".join(self._get_data(f'https://openml.org/api/v1/json/data/{data_id}', self._cache_keys['data']))
        descr_obj = json.loads(descr_txt)["data_set_description"]
        return descr_obj

    def _get_feat_descr(self, data_id:int) -> Sequence[Dict[str,Any]]:
        descr_txt = " ".join(self._get_data(f'https://openml.org/api/v1/json/data/features/{data_id}', self._cache_keys['feat']))
        descr_obj = json.loads(descr_txt)["data_features"]["feature"]
        return descr_obj

    def _get_task_descr(self, task_id) -> Dict[str,Any]:
        descr_txt = " ".join(self._get_data(f'https://openml.org/api/v1/json/task/{task_id}', self._cache_keys['task']))
        descr_obj = json.loads(descr_txt)['task']

        task_type   = int(descr_obj.get('task_type_id',0))
        task_source = ([i for i in descr_obj.get("input",[]) if i.get('name',None) == 'source_data'] + [{}])[0]
        data_id     = int(task_source.get("data_set",{}).get("data_set_id",0)) or None
        target      = task_source.get("data_set",{}).get("target_feature","") or None

        return { 'id': task_id, 'type': task_type, 'data': data_id, 'target': target}

    def _get_arff_lines(self, file_id:str, md5_checksum:str) -> Iterable[str]:
            arff_url = f"https://openml.org/data/v1/download/{file_id}"
            arff_key = self._cache_keys['arff']
            return self._get_data(arff_url, arff_key, md5_checksum)

    def _clear_cache(self) -> None:
        for key in self._cache_keys.values():
            CobaContext.cacher.rmv(key)

    def _source_already_cached(self) -> bool:
        old_data_id = self._data_id
        all_cached  = False

        if self._task_id and self._cache_keys['task'] in CobaContext.cacher:
            task_descr = self._get_task_descr(self._task_id)

            if not task_descr['data']: return True #this will fall into an exception so no more caching is needed

            self._data_id = task_descr['data']

        all_cached = bool(self._data_id and all(self._cache_keys[k] in CobaContext.cacher for k in ['data','feat','arff']))

        self._data_id = old_data_id
        return all_cached

    @property
    def _cache_keys(self):
        return {
            'task': f"openml_{self._task_id or 0:0>6}_task",
            'data': f"openml_{self._data_id or 0:0>6}_data",
            'feat': f"openml_{self._data_id or 0:0>6}_feat",
            'arff': f"openml_{self._data_id or 0:0>6}_arff",
        }

class OpenmlSimulation(SupervisedSimulation):
    """A supervised simulation created from an openml dataset.
    Download a dataset from openml.org and create a SupervisedSimulation.
    """

    @overload
    def __init__(self, data_id: int, drop_missing: bool = True, take: int = None, *, target:str = None, ):
        """Instantiate an OpenmlSimulation.
        Args:
            data_id: The data id uniquely identifying the dataset on openml (i.e., openml.org/d/{id})
            drop_missing: Drop data rows with missing values.
            take: The number of interactions we'd like the simulation to have (these will be selected at random).
            target: The column that should be marked as the prediction label in the source.
        """
        ...

    @overload
    def __init__(self, *, task_id: int, drop_missing: bool = True, take: int=None, target:str = None):
        """Instantiate an OpenmlSimulation.
        Args:
            task_id: The openml task id which identifies the dataset to use from openml along with its label
            drop_missing: Drop data rows with missing values.
            take: The number of interactions we'd like the simulation to have (these will be selected at random).
            target: The column that should be marked as the prediction label in the source.
        """
        ...

    def __init__(self, *args, **kwargs) -> None:
        """Instantiate an OpenmlSimulation."""
        kwargs.update(zip(['data_id','drop_missing','take'], args))
        super().__init__(OpenmlSource(**kwargs), None, kwargs.get('label_type',None), kwargs.get('take',None))

    @property
    def params(self) -> Dict[str, Any]:
        return {**super().params, "env_type": "OpenmlSimulation" }

    def __str__(self) -> str:

        params = [
            f"data={self.params['openml_data']}" if self.params.get('openml_data') else f"task={self.params['openml_task']}",
            f"target={self.params['openml_target']}" if self.params.get('openml_target') else "",
        ]

        return f"Openml({str.join(', ', filter(None,params))})"



import json
import collections.abc

from itertools import product
from typing import Sequence, Iterable, Union, Dict, Any

from coba.context import CobaContext
from coba.registry import JsonMakerV1, CobaRegistry, JsonMakerV2
from coba.pipes import UrlSource, Pipes
from coba.exceptions import CobaException
from coba.primitives import Source, Environment

class EnvironmentsTemplateV1(Source[Sequence[Environment]]):

    def __init__(self, source: Union[Source[Iterable[str]],str]) -> None:
        self._source = UrlSource(source) if isinstance(source,str) else source

    def read(self) -> Sequence[Environment]:
        definitions: dict = json.loads('\n'.join(self._source.read()))
        variables = { k: self._construct_via_method(v) for k,v in definitions.get("variables",{}).items() }

        def _construct(item:Any) -> Sequence[Any]:
            result = None

            if isinstance(item, str) and item in variables:
                result = variables[item]

            if isinstance(item, str) and item not in variables:
                result = self._construct_via_method(item)

            if isinstance(item, dict):
                result = self._construct_via_method(item)

            if isinstance(item, list):
                pieces = list(map(_construct, item))

                if hasattr(pieces[0][0],'read'):
                    result = [ Pipes.join(s,*f) for s in pieces[0] for f in product(*pieces[1:])]
                else:
                    result = sum(pieces,[])

            if result is None:
                raise CobaException(f"We were unable to construct {item} in the given environment definition file.")

            return result if isinstance(result, collections.abc.Sequence) else [result]

        if not isinstance(definitions['environments'], list): definitions['environments'] = [definitions['environments']]

        return [ environment for recipe in definitions['environments'] for environment in _construct(recipe)]

    def _construct_via_method(self, item):
        return JsonMakerV1(CobaRegistry.registry).make(item)

class EnvironmentsTemplateV2(Source[Sequence[Environment]]):

    def __init__(self, source: Union[str,Source[Iterable[str]]], **user_vars) -> None:
        self._source = UrlSource(source) if isinstance(source,str) else source
        self._user_variables = { f"${k}" if k[0] != "$" else k :v for k,v in user_vars.items() }

    def read(self) -> Sequence[Environment]:
        definition: dict = json.loads('\n'.join(self._source.read()))

        needed_var = list(self._missing(definition.get("variables",{}))) + list(self._missing(definition.get('environments',[])))
        not_needed_user_var = [ v[1:] for v in set(self._user_variables.keys()) - set(needed_var)]

        if not_needed_user_var:
            CobaContext.logger.log(f"The following provided variables were not used by the template:{not_needed_user_var}")

        old_variables = { **definition.get("variables",{}), **self._user_variables}
        new_variables = self._fill(old_variables,old_variables)
        num_fillings  = 0

        while old_variables != new_variables:
            num_fillings += 1
            old_variables = new_variables
            new_variables = self._fill(old_variables,old_variables)

            if num_fillings > 4:
                raise CobaException("There appears to be an infinite loop in the template.")

        variables = { k: JsonMakerV2(CobaRegistry.registry).make(v,strict=False) for k,v in new_variables.items() }

        recipes = definition['environments']
        recipes = recipes if isinstance(recipes, list) else [recipes]
        recipes = self._fill(recipes, variables)

        missing = set(self._missing(recipes))
        if missing: raise CobaException(f"The following variables were not defined: [{','.join(missing)}]")

        environments = []
        for recipe in recipes:
            environments.extend(self._make(recipe))

        environments = sorted(environments, key= lambda env: env.params.get("shuffle_seed",0))

        return environments

    def _make(self, item:Union[str,list,dict] ) -> Sequence[Any]:

        result = item

        if isinstance(item, str):
            result = JsonMakerV2().make(item)

        if isinstance(item, dict):
            result = JsonMakerV2().make(item)

        if isinstance(item, list):
            pieces = [ self._make(i) for i in item]

            if hasattr(pieces[0][0],'read') and hasattr(pieces[1][0],'filter'):
                result = [ Pipes.join(s,*f) for s in pieces[0] for f in product(*pieces[1:])]
            else:
                result = sum(pieces,[])

        if result is None:
            raise CobaException(f"We were unable to construct {item} in the given environment definition file.")

        return result if isinstance(result, collections.abc.Sequence) else [result]

    def _fill(self, target: Any, variables:Dict[str,Any]):
        if isinstance(target,dict):
            return { k: self._fill(v,variables) for k,v in target.items() }
        if isinstance(target,list):
            return [ self._fill(i,variables) for i in target ]
        if isinstance(target,str) and target in variables:
            return variables[target]
        return target

    def _missing(self, target: Any) -> Sequence[str]:
        if isinstance(target,dict):
            return sum([self._missing(v) for v in target.values()],[])
        if isinstance(target,list):
            return sum([self._missing(i) for i in target],[])
        if isinstance(target,str) and len(target) > 1 and target.startswith("$") and not target[1:].isdigit():
            return [target]
        return []



from typing import Any, Mapping, Sequence, Tuple

from coba.exceptions import CobaException
from coba.utilities import PackageChecker
from coba.primitives import Learner, Context, Action, Actions, Prob
from coba.encodings import InteractionsEncoder
from coba.learners.utilities import PMFPredictor

class LinTSLearner(Learner):
    """A contextual bandit learner using Thompson Sampling for exploration.

    This is an implementation of the Agrawal et al. (2013) Thompson Sapmling
    algorithm. The `Sherman-Morrison formula`__ is utilized to iteratively
    calculate the inversion matrix. Expected reward is represented as a
    linear function of context and action features.

    Remarks:
        A small note on the stability of the Sherman-Morrison formula can be found `here`__.

    References:
        Agrawal, Shipra, and Navin Goyal. "Thompson sampling for contextual bandits with
        linear payoffs." International conference on machine learning. PMLR, 2013.

    __ https://en.wikipedia.org/wiki/Sherman%E2%80%93Morrison_formula
    __ https://scicomp.stackexchange.com/q/20386/46891
    """

    def __init__(self, v: float = 1, features: Sequence[str] = [1, 'a', 'ax'], seed: int = 1) -> None:
        """Instantiate a LinUCBLearner.

        Args:
            v: Modify the exploration rate of the algorithm. A value of 0 will not explore
                while a value of `inf` will explores uniformly forever. The appropriate setting
                of v will depend to some degree on the scale of given feature vectors and rewards.
            features: Feature set interactions to use when calculating action value estimates.
                Context features are indicated by x's while action features are indicated by
                a's. For example, xaa means to cross context and action and action features.
            seed: A seed for a random number generation.
        """
        PackageChecker.numpy("LinTSLearner")

        self._X         = features
        self._X_encoder = InteractionsEncoder(features)

        self._B_inv  = None
        self._mu_hat = None
        self._v      = v
        self._pred   = PMFPredictor(self._pmf,seed)

    @property
    def params(self) -> Mapping[str, Any]:
        return {'family': 'LinTS', 'v': self._v, 'features': self._X, 'seed': self._pred.seed}

    def _initialize(self, context, action) -> None:
        if isinstance(action, dict) or isinstance(context, dict):
            raise CobaException("Sparse data cannot be handled by this implementation at this time.")

        if not context:
            self._X_encoder = InteractionsEncoder(list(set(filter(None,[ f.replace('x','') if isinstance(f,str) else f for f in self._X ]))))

        d  = len(self._X_encoder.encode(x=context or [],a=action))
        np = __import__('numpy')

        self._np     = np
        self._nrng   = np.random.default_rng(1)
        self._mu_hat = np.zeros(d)
        self._B_inv  = np.identity(d)

    def _pmf(self, context, actions):
        if self._B_inv is None: self._initialize(context,actions[0])

        np = self._np

        context  = context or []
        features = np.array([self._X_encoder.encode(x=context,a=action) for action in actions])

        if self._v == 0:
            mu_tilde = self._mu_hat
        else:
            mu_tilde = self._nrng.multivariate_normal(self._mu_hat, self._v*self._B_inv, method='cholesky')

        point_estimates = mu_tilde @ features.T
        max_indexes     = np.where(point_estimates.round(5) == np.amax(point_estimates).round(5))[0]

        return [int(ind in max_indexes)/len(max_indexes) for ind in range(len(actions))]

    def score(self, context: 'Context', actions: 'Actions', action: 'Action') -> 'Prob':
        return self._pred.score(context,actions,action)

    def predict(self, context: 'Context', actions: 'Actions') -> Tuple['Action','Prob']:
        return self._pred.predict(context,actions)

    def learn(self, context: 'Context', action: 'Action', reward: float, probability: float) -> None:
        if self._B_inv is None: self._initialize(context,action)

        np = self._np

        context  = context or []
        features = np.array(self._X_encoder.encode(x=context,a=action)).T

        r = self._mu_hat @ features
        w = self._B_inv  @ features
        v = w            @ features

        self._mu_hat = self._mu_hat + (reward-r)/(1+v)*w
        self._B_inv  = self._B_inv - np.outer(w,w)/(1+v)



"""This module contains utility classes for transforming data between encodings."""

import time

from numbers import Number
from abc import ABC, abstractmethod
from collections import Counter, OrderedDict, defaultdict
from itertools import count, accumulate, chain
from typing import Iterator, Sequence, Generic, TypeVar, Any, Tuple, Union, Mapping

from coba.exceptions import CobaException
from coba.primitives import Sparse, Dense, Categorical

_T_out = TypeVar('_T_out', bound=Any, covariant=True)

class Encoder(Generic[_T_out], ABC):
    """The Encoder interface.

    Remarks:
        While it can't be enforced by the interface, the assumption is that all Encoder
        implementations are immutable. This means that the `fit` method should always
        return a new Encoder.
    """

    @property
    @abstractmethod
    def is_fit(self) -> bool:
        """Indicates if the encoder has been fit.

        Returns:
            True if the Encoder is ready to encode otherwise False.
        """
        ...

    @abstractmethod
    def fit(self, values: Sequence[Any]) -> 'Encoder':
        """Fit the encoder given training data.

        Args:
            values: The encoder training values.

        Returns:
            A fit Encoder.
        """
        ...

    @abstractmethod
    def encode(self, value: Any) -> Sequence[_T_out]:
        """Encode the given value.

        Args:
            value: The value to encode.

        Returns:
            The encoded value.
        """
        ...

    @abstractmethod
    def encodes(self, values: Sequence[Any]) -> Sequence[_T_out]:
        """Encode the given values.

        Args:
            values: The values to encode.

        Returns:
            The encoded values.
        """
        ...

    def fit_encodes(self, values: Sequence[Any]) -> Sequence[_T_out]:
        """Fit and then encode the given values.

        Args:
            values: The values to encode.

        Returns:
            The encoded values.
        """
        if self.is_fit:
            return self.encodes(values)
        else:
            return self.fit(values).encodes(values)

class IdentityEncoder(Encoder[Any]):

    @property
    def is_fit(self) -> bool:
        return True

    def fit(self, values: Sequence[Any]) -> 'Encoder':
        return self

    def encode(self, value: Any) -> Any:
        return value

    def encodes(self, values: Sequence[Any]) -> Sequence[Any]:
        return values

class MissingEncoder(Encoder[Any]):

    def __init__(self, encoder:Encoder = IdentityEncoder(), missing_vals: Sequence[Any] = ["?",""], missing_rep: Any = None) -> None:
        self._encoder      = encoder
        self._missing_vals = missing_vals
        self._missing_rep  = missing_rep

    @property
    def is_fit(self) -> bool:
        return True

    def fit(self, values: Sequence[Any]) -> 'Encoder':
        return self

    def encode(self, value: Any) -> Any:
        return self._missing_rep if value in self._missing_vals else self._encoder.encode(value)

    def encodes(self, values: Sequence[Any]) -> Sequence[Any]:
        return [ self.encode(v) for v in values ]

class StringEncoder(Encoder[str]):
    """An Encoder implementation that turns incoming values into string values."""

    @property
    def is_fit(self) -> bool:
        return True

    def fit(self, values: Sequence[Any]) -> 'StringEncoder':
        return StringEncoder()

    def encode(self, value: Any) -> str:
        return str(value)

    def encodes(self, values: Sequence[Any]) -> Sequence[str]:
        return list(map(str,values))

class NumericEncoder(Encoder[float]):
    """An Encoder implementation that turns incoming values into float values."""

    @property
    def is_fit(self) -> bool:
        return True

    def fit(self, values: Sequence[Any]) -> 'NumericEncoder':
        return NumericEncoder()

    def encode(self, value: Any) -> float:
        try:
            return float(value)
        except:
            return float('nan')

    def encodes(self, values: Sequence[Any]) -> Sequence[float]:
        return list(self._float_generator(values))

    def _float_generator(self,values) -> Iterator[float]:
        for value in values:
            try:
                yield float(value)
            except:
                yield float('nan')

class OneHotEncoder(Encoder[Tuple[int,...]]):
    """An Encoder implementation that turns incoming values into a one hot representation."""

    def __init__(self, values: Sequence[Any] = [], err_if_unknown = False) -> None:
        """Instantiate a OneHotEncoder.

        Args:
            values: Provide the universe of values for encoding and set `is_fit==True`.
            err_if_unknown: When an unknown value is passed to `encode` throw an exception (otherwise encode as all 0's).
        """

        self._err_if_unknown = err_if_unknown
        self._onehots        = None
        self._default        = None

        if values:

            values = sorted(set(values), key=lambda v: values.index(v))

            self._default = tuple([0] * len(values))
            known_onehots = [ [0] * len(values) for _ in range(len(values)) ]

            for i,k in enumerate(known_onehots):
                k[i] = 1

            keys_and_values = zip(values, map(tuple,known_onehots))

            if self._err_if_unknown:
                self._onehots = dict(keys_and_values)
            else:
                self._onehots = defaultdict(lambda:self._default, keys_and_values)

    @property
    def is_fit(self) -> bool:
        return self._onehots is not None

    def fit(self, values: Sequence[Any]) -> 'OneHotEncoder':
        return OneHotEncoder(values = values, err_if_unknown = self._err_if_unknown)

    def encode(self, value: Any) -> Tuple[int,...]:
        if self._onehots is None:
            raise CobaException("This encoder must be fit before it can be used.")

        try:
            return self._onehots[value]
        except KeyError as e:
            raise CobaException(f"We were unable to find {e} in {list(self._onehots.keys())}")

    def encodes(self, values: Sequence[Any]) -> Sequence[Tuple[int,...]]:
        if self._onehots is None:
            raise CobaException("This encoder must be fit before it can be used.")

        try:
            return list(map(self._onehots.__getitem__,values))
        except KeyError as e:
            raise CobaException(f"We were unable to find {e} in {list(self._onehots.keys())}")

class CategoricalEncoder(Encoder[Categorical]):
    """An Encoder implementation that turns incoming values into a one hot representation."""

    def __init__(self, values: Sequence[str] = []) -> None:
        """Instantiate a OneHotEncoder.

        Args:
            values: Provide the universe of values for encoding and set `is_fit==True`.
        """

        set_values = set(values)
        if len(values) != len(set_values): values = sorted(set_values)
        self._categoricals = {v: Categorical(v,values) for v in sorted(set(values)) } if values else None

    @property
    def is_fit(self) -> bool:
        return self._categoricals is not None

    def fit(self, values: Sequence[str]) -> 'CategoricalEncoder':
        return CategoricalEncoder(values = values)

    def encode(self, value: Any) -> Categorical:
        try:
            return self._categoricals[value]
        except KeyError as e:
            raise CobaException(f"We were unable to find {e} in the categorical values of {list(self._categoricals.keys())}.")
        except TypeError:
            raise CobaException("This encoder must be fit before it can be used.")

    def encodes(self, values: Sequence[Any]) -> Sequence[Tuple[int,...]]:
        try:
            return list(map(self._categoricals.__getitem__,values))
        except KeyError as e:
            raise CobaException(f"We were unable to find {e} in the categorical values of {list(self._categoricals.keys())}.")
        except (TypeError, AttributeError):
            raise CobaException("This encoder must be fit before it can be used.")

class FactorEncoder(Encoder[int]):
    """An Encoder implementation that turns incoming values into factor representation."""

    def __init__(self, values: Sequence[Any] = [], err_if_unknown = False) -> None:
        """Instantiate a FactorEncoder.

        Args:
            values: Provide the universe of values for encoding and set `is_fit==True`.
            err_if_unknown: When an unknown value is passed to `encode` throw an exception (otherwise encode as all 0's).
        """

        self._err_if_unknown = err_if_unknown
        self._levels         = None

        if values:
            values = sorted(set(values), key=lambda v: values.index(v))
            levels  = [ i + 1 for i in range(len(values)) ]

            pairs = zip(values, levels)
            self._levels = dict(pairs) if self._err_if_unknown else defaultdict(lambda: float('nan'), pairs)

    @property
    def is_fit(self) -> bool:
        return self._levels is not None

    def fit(self, values: Sequence[Any]) -> 'FactorEncoder':
        return FactorEncoder(values = values, err_if_unknown = self._err_if_unknown)

    def encode(self, value: Any) -> int:
        if self._levels is None:
            raise CobaException("This encoder must be fit before it can be used.")

        try:
            return self._levels[value]
        except KeyError as e:
            raise CobaException(f"We were unable to find {e} in {self._levels.keys()}") from None

    def encodes(self, values: Sequence[Any]) -> Sequence[int]:
        if self._levels is None:
            raise CobaException("This encoder must be fit before it can be used.")

        try:
            return list(map(self._levels.__getitem__,values))
        except KeyError as e:
            raise CobaException(f"We were unable to find {e} in {self._levels.keys()}") from None

class InteractionsEncoder:

    def __init__(self, interactions: Sequence[Union[str,float]]) -> None:
        str_interactions = [i for i in interactions if isinstance(i,str)   ]
        num_interactions = [i for i in interactions if isinstance(i,Number)]

        self.times       = [0,0,0,0]
        self.n           = 0
        self._constant   = sum(num_interactions)
        self._cross_pows = OrderedDict(zip(interactions,map(OrderedDict,map(Counter,str_interactions))))
        self._ns_max_pow = { n:int(max(p.get(n,0) for p in self._cross_pows.values())) for n in set(''.join(str_interactions)) }

    def encode(self, **ns_raw_values: Union[str, float, Sequence[Union[str,float]], Mapping[Union[str,int],Union[str,float]]]) -> Union[Sequence[float], Mapping[str,float]]:
        self.n+= 1

        ns_raw_values = { k:v if v is not None else [] for k,v in ns_raw_values.items() }

        is_str = lambda v: isinstance(v,str)
        is_seq = lambda v: isinstance(v,Dense)
        is_map = lambda v: isinstance(v,Sparse)

        is_sparse_type = lambda f: is_map(f) or is_str(f)
        is_sparse_sequ = lambda f: is_seq(f) and any(map(is_sparse_type,f))

        is_sparse = any(is_sparse_type(v) or is_sparse_sequ(v) for v in ns_raw_values.values())

        def make_dict(v) -> Mapping[str,Union[str,float]]:
            return v if is_map(v) else dict(zip(map(str,count()),v)) if is_seq(v) else { "0":v }

        def make_list(v) -> Sequence[Union[str,float]]:
            return v if is_seq(v) else [v]

        def handle_str(v: Mapping[str,Union[str,float]]) -> Mapping[str,float]:
            return { (f"{x}{y}" if is_str(y) else x):(1 if is_str(y) else y) for x,y in v.items() }

        start = time.time()
        if is_sparse:
            ns_values = { ns:handle_str(make_dict(V))            for ns,V in ns_raw_values.items() if ns in self._ns_max_pow }
            ns_values = { ns:{f"{ns}{k}":v for k,v in V.items()} for ns,V in ns_values.items()     if ns in self._ns_max_pow }
        else:
            ns_values = { ns:make_list(v) for ns,v in ns_raw_values.items() if ns in self._ns_max_pow}
        self.times[0] += time.time()-start

        pows = self._pows
        cross = self._cross

        if is_sparse:
            start = time.time()
            key_pows = { ns: pows(list(ns_values[ns].keys()  ), max_pow) for ns, max_pow in self._ns_max_pow.items() }
            val_pows = { ns: pows(list(ns_values[ns].values()), max_pow) for ns, max_pow in self._ns_max_pow.items() }
            self.times[1] += time.time()-start

            start = time.time()
            key_crosses = [ cross(key_pows, cross_pow) for cross_pow in self._cross_pows.values() ]
            val_crosses = [ cross(val_pows, cross_pow) for cross_pow in self._cross_pows.values() ]
            self.times[2] += time.time()-start

            start = time.time()
            encoded = dict(zip(chain.from_iterable(key_crosses), chain.from_iterable(val_crosses)))
            self.times[3] += time.time()-start

            if self._constant: encoded['const'] = self._constant

            return encoded

        else:
            start = time.time()
            val_pows = { ns: pows(ns_values[ns], max_pow) for ns, max_pow in self._ns_max_pow.items() }
            self.times[1] += time.time()-start

            start = time.time()
            val_crosses = [ cross(val_pows, cross_pow) for cross_pow in self._cross_pows.values() ]
            self.times[2] += time.time()-start

            start = time.time()
            encoded = sum(val_crosses,[])
            self.times[3] += time.time()-start

            if self._constant: encoded = [self._constant] + encoded

            return encoded

    def _pows(self, values: Sequence[Union[str,float]], degree):
        #WARNING: This function has been extremely optimized. Please baseline performance before and after making any changes.
        #WARNING: You can find three existing performance tests in test_performance.
        if not values: return []

        starts = [1]*len(values)
        terms  = [['']] if isinstance(values[0],str) else [[1]]

        for d in range(degree):
            if isinstance(values[0],str):
                terms.append([v+t for v,s in zip(values,starts) for t in terms[d][(s-1):]])
            else:
                terms.append([v*t for v,s in zip(values,starts) for t in terms[d][(s-1):]])

            starts = list(accumulate(starts[:1]+starts[-1:]+starts[1:-1]))

        return terms

    def _cross(self, ns_pows, cross_pow):
        #WARNING: This function has been extremely optimized. Test speed before and after making any changes.
        #WARNING: Look in test_performance for three existing performance tests.
        if any(ns_pows[k] == [] for k in cross_pow.keys()): return []

        values = [ ns_pows[ns][p] for ns,p in cross_pow.items() ]
        cross  = values[0]

        if isinstance(cross[0],str):
            for vs in values[1:]: cross = [ o+v for o in cross for v in vs ]
        else:
            for vs in values[1:]: cross = [ o*v for o in cross for v in vs ]

        return cross
